# coding=utf-8
"""
Classes for generating Events from stochastic models.
"""
from typing import List, Dict, Any, Callable, Tuple, Type, Optional
from acnportal.acnsim.events import EventQueue, PluginEvent
from acnportal.acnsim.models import EV, Battery
import numpy as np
import pandas as pd
from typing_extensions import TypedDict

try:
    from sklearn.mixture import GaussianMixture
except ImportError as e:
    raise ImportError(
        "scikit-learn is required to use the stochastic_events module. "
        "Please install scikit-learn using\n"
        "\tpip install scikit-learn\n"
        "or install acnportal with extra options\n"
        "\tpip install acnportal[scikit-learn]"
    )


CapFnCallable = Callable[[float, float, float, float], Tuple[float, float]]

BatteryParams = TypedDict(
    "BatteryParams",
    {"type": Type[Battery], "capacity_fn": CapFnCallable, "kwargs": Dict[str, Any]},
    total=False,
)


class StochasticEvents:
    """ Base class for generating events from a stochastic model.

    Args:
        arrival_min (float): Clip arrive time at lower bound. Useful if you know that drivers should not arrive before a
            certain time. By default 0, meaning drivers do not arrive on the previous day. [hours since midnight]
        arrival_max (float): Clip arrive time at upper bound. Useful if you know drivers should not arrive after a
            certain time. My default 24, meaning drivers do not arrive the next day. [hours since midnight]
        duration_min (float): Clip duration to lower bound. Useful to limit session duration to a reasonable minimum.
            By default 0.0833 hours or 5 minutes. [hours]
        duration_max (float): Clip duration to upper bound. Useful to limit session duration to a reasonable maximum.
            By default 48 hours. [hours]
        energy_min (float): Clip energy request to lower bound. Useful to limit energy request to a reasonable minimum.
            By default 0.5 kWh. [kWh]
        energy_max (float): Clip energy request to upper bound. Useful to limit energy request to a reasonable maximum.
            By default 150 kWh. [kWh]
    """

    arrival_min: float
    arrival_max: float
    duration_min: float
    duration_max: float
    energy_min: float
    energy_max: float

    def __init__(
        self,
        arrival_min: float = 0.0,
        arrival_max: float = 24.0,
        duration_min: float = 0.0833,
        duration_max: float = 48.0,
        energy_min: float = 0.5,
        energy_max: float = 150.0,
    ) -> None:
        self.arrival_min = arrival_min
        self.arrival_max = arrival_max
        self.duration_min = duration_min
        self.duration_max = duration_max
        self.energy_min = energy_min
        self.energy_max = energy_max

    def fit(self, data: List[Dict[str, Any]], **kwargs) -> None:
        """ Fit StochasticEvents model to data from ACN-Data.

        Args:
            data (List[Dict[str, Any]]): List of session dictionaries.
                See DataClient.get_sessions().

        Returns:
            None
        """

    def sample(self, n_samples: int) -> np.ndarray:
        """ Generate random samples from the fitted model.

        Args:
            n_samples (int): Number of samples to generate.

        Returns:
            np.ndarray: shape (n_samples, 3), randomly generated samples. Column 1 is
                the arrival time in hours since midnight, column 2 is the duration of
                the stay in hours, and column 3 is the energy demand in kWh.
        """
        pass

    def clip_samples(self, sample_matrix: np.ndarray) -> np.ndarray:
        """ Clip samples matrix into their upper and lower bounds.

            Note that this function will modify the samples_matrix in place as well as
            return it.

        Args:
            sample_matrix (np.ndarray): shape (n_samples, 3), randomly generated
                samples. Column 1 is the arrival time in hours since midnight,
                column 2 is the session duration time in hours, and column 3 is the energy
                demand in kWh.

        Returns:
            np.ndarray: sample matrix with all entries projected into their
                corresponding bounds.
        """
        sample_matrix[:, 0] = np.clip(
            sample_matrix[:, 0], self.arrival_min, self.arrival_max
        )
        sample_matrix[:, 1] = np.clip(
            sample_matrix[:, 1], self.duration_min, self.duration_max
        )
        sample_matrix[:, 2] = np.clip(
            sample_matrix[:, 2], self.energy_min, self.energy_max
        )
        return sample_matrix

    def generate_events(
        self,
        sessions_per_day: List[int],
        period: float,
        voltage: float,
        max_battery_power: float,
        max_len: int = None,
        battery_params: Optional[BatteryParams] = None,
        force_feasible: bool = False,
    ) -> EventQueue:
        """ Return EventQueue from random generated samples.

            Args:
                sessions_per_day (List[int]): Number of sessions to sample for each day
                    of the simulation.
                period (int): Length of each time interval. (minutes)
                voltage (float): Voltage of the network.
                max_battery_power (float): Default maximum charging power for batteries.
                max_len (int): Maximum length of a session. (periods) Default None.
                battery_params (Dict[str, object]): Dictionary containing parameters for
                    the EV's battery. Three keys are supported. If none, Battery type
                    is used with default configuration. Default None.
                        - 'type' maps to a Battery-like class. (required)
                        - 'capacity_fn' maps to a function which takes in the the energy
                            delivered to the car, the length of the session,
                            the period of the simulation, and the voltage of the
                            system. It should return a tuple with the capacity of the
                            battery and the initial charge of the battery both in
                            A*periods.
                        - 'kwargs' maps to a dictionary of keyword arguments which will
                            be based to the Battery constructor.
                force_feasible (bool): If True, the requested_energy of each session
                    will be reduced if it exceeds the amount of energy which could be
                    delivered at maximum rate during the duration of the charging
                    session. Default False. Returns: EventQueue: Queue of plugin
                    events for the samples charging sessions.
        """
        daily_sessions: List[np.ndarray] = []
        for d, num_sessions in enumerate(sessions_per_day):
            if num_sessions > 0:
                daily_arrivals = self.sample(num_sessions)
                daily_arrivals[:, 0] += 24 * d
                daily_sessions.append(daily_arrivals)
        ev_matrix = np.vstack([day for day in daily_sessions if day is not None])
        evs = self._convert_ev_matrix(
            ev_matrix,
            period,
            voltage,
            max_battery_power,
            max_len,
            battery_params,
            force_feasible,
        )
        events = [PluginEvent(sess.arrival, sess) for sess in evs]
        return EventQueue(events)

    @staticmethod
    def extract_training_data(data: List[Dict[str, Any]]):
        """ Generate matrix for training Gaussian Mixture Model.

        Args:
            data (List[Dict[str, Any]]): List of session dictionaries.
                See DataClient.get_sessions().

        Returns:
            np.ndarray: shape(n_sessions, 3) Column 1 is the arrival time in hours since
                midnight, column 2 is the session duration in hours, and column 3 is the
                energy demand in kWh.
        """
        df = pd.DataFrame(data)
        df.sort_values(by="connectionTime", inplace=True)
        connection_time = [v.hour + v.minute / 60 for v in df["connectionTime"]]
        durations = [
            v.total_seconds() / 3600
            for v in df["disconnectTime"] - df["connectionTime"]
        ]
        energy = [v for v in df["kWhDelivered"]]
        return np.array([connection_time, durations, energy]).T

    @staticmethod
    def _convert_ev_matrix(
        ev_matrix: np.ndarray,
        period: float,
        voltage: float,
        max_battery_power: float,
        max_len: int = None,
        battery_params: Optional[BatteryParams] = None,
        force_feasible: bool = False,
    ) -> List[EV]:
        """

        Args:
            ev_matrix (np.ndarray[float]): Nx3 array where N is the number of EVs.
                Column 1 is the arrival time in hours since midnight, column 2 is the
                session duration in hours, and column 3 is the energy demand in kWh.
            (See generate_events() for other arguments)

        Returns:
            List[EV]: List of EVs with parameters taken from ev_matrix.
        """

        period_per_hour = 60 / period
        evs = []
        for row_idx, row in enumerate(ev_matrix):
            arrival, duration, energy_delivered = row

            if arrival < 0 or duration <= 0 or energy_delivered <= 0:
                print("Invalid session.")
                continue

            if max_len is not None and duration > max_len:
                duration = max_len

            if force_feasible:
                max_feasible = max_battery_power * duration
                energy_delivered = np.minimum(max_feasible, energy_delivered)

            departure = int((arrival + duration) * period_per_hour)
            arrival = int(arrival * period_per_hour)
            session_id = f"session_{row_idx}"
            # By default a new station is created for each EV.
            # Infinite space assumption.
            station_id = f"station_{row_idx}"

            battery_params_input: BatteryParams
            if battery_params is None:
                battery_params_input = {"type": Battery}
            else:
                battery_params_input = battery_params
            battery_kwargs = (
                battery_params_input["kwargs"]
                if "kwargs" in battery_params_input
                else {}
            )
            if "capacity_fn" in battery_params_input:
                cap_fn: CapFnCallable = battery_params_input["capacity_fn"]
                cap, init = cap_fn(energy_delivered, duration, voltage, period)
            else:
                cap = energy_delivered
                init = 0
            battery_type = battery_params_input["type"]
            battery = battery_type(cap, init, max_battery_power, **battery_kwargs)
            evs.append(
                EV(
                    arrival,
                    departure,
                    energy_delivered,
                    station_id,
                    session_id,
                    battery,
                )
            )
        return evs


class GaussianMixtureEvents(StochasticEvents):
    """ Model to draw charging session parameters from a gaussian mixture model.

    Args:
        pretrained_model (GaussianMixture): A trained Gaussian Mixture Model with
            variables arrival time (h), session duration (h), energy demand (kWh).
        Also accepts any kwargs for the sklearn GaussianMixture class.
        See https://scikit-learn.org/stable/modules/generated/sklearn.mixture.GaussianMixture.html.
    """

    gmm: GaussianMixture

    def __init__(
        self,
        arrival_min: float = 0.0,
        arrival_max: float = 24.0,
        duration_min: float = 0.0833,
        duration_max: float = 48.0,
        energy_min: float = 0.5,
        energy_max: float = 150.0,
        pretrained_model: Optional[GaussianMixture] = None,
        **kwargs,
    ):
        super().__init__(
            arrival_min, arrival_max, duration_min, duration_max, energy_min, energy_max
        )
        if pretrained_model is None:
            self.gmm = GaussianMixture(**kwargs)
        else:
            self.gmm = pretrained_model

    def fit(self, data: List[Dict[str, Any]], **kwargs) -> None:
        """ Fit StochasticEvents model to data from ACN-Data.

        Args:
            data (List[Dict[str, Any]]): List of session dictionaries.
            See DataClient.get_sessions().
            Also accepts any kwargs for the sklearn GaussianMixture class fit method.

        Returns:
            None
        """
        x = self.extract_training_data(data)
        self.gmm.fit(x, **kwargs)

    def sample(self, n_samples: int):
        """ Generate random samples from the fitted model.

        Args:
            n_samples (int): Number of samples to generate.

        Returns:
            np.ndarray: shape (n_samples, 3), randomly generated samples. Column 1 is
                the arrival time in hours since midnight, column 2 is the session duration in hours,
                and column 3 is the energy demand in kWh.
        """
        if n_samples > 0:
            ev_matrix, _ = self.gmm.sample(n_samples)
            return self.clip_samples(ev_matrix)
        else:
            return np.array([])



import random
from collections import OrderedDict
import warnings

from acnportal.acnsim.network.charging_network import ChargingNetwork
from acnportal.acnsim.models import EV


class StochasticNetwork(ChargingNetwork):
    def __init__(
        self,
        violation_tolerance: float = 1e-5,
        relative_tolerance: float = 1e-7,
        early_departure: bool = False,
    ):
        """ Extends ChargingNetwork to support non-deterministic space assignment."""
        super().__init__(violation_tolerance, relative_tolerance)
        self.waiting_queue = OrderedDict()
        self.early_departure = early_departure

        # Stats
        self.swaps = 0
        self.never_charged = 0
        self.early_unplug = 0

    def available_evses(self):
        """ Return a list of all EVSEs which do not have an EV attached. """
        return [evse_id for evse_id, evse in self._EVSEs.items() if evse.ev is None]

    def plugin(self, ev, station_id=None):
        """ Attaches EV to a random EVSE if one is available, otherwise places EV in the
            waiting queue.

        Args:
            ev (EV): EV object which will be attached to the EVSE.
            [Depreciated]
            station_id (str): ID of the EVSE.

        Returns:
            None

        Raises:
            KeyError: Raised when the station id has not yet been registered.
        """
        if station_id is not None:
            warnings.warn(
                "plugin arg 'station_id' is deprecated. Plugin now uses the value of ev.station_id directly.",
                DeprecationWarning,
            )
        available_spots = self.available_evses()
        if len(available_spots) > 0:  # and len(self.waiting_queue) == 0:
            chosen_spot = random.choice(available_spots)
            ev.update_station_id(chosen_spot)
            super().plugin(ev)
        else:
            ev.update_station_id(None)
            self.waiting_queue[ev.session_id] = ev
            self.waiting_queue.move_to_end(ev.session_id)

    def unplug(self, station_id: str, session_id: str = None) -> None:
        """ Detach EV from a specific EVSE.

        Args:
            station_id (str): ID of the EVSE.
            session_id (str): ID of the session to be unplugged.
        Returns:
            None

        Raises:
            KeyError: Raised when the station id has not yet been registered.
        """
        if session_id in self.waiting_queue:
            del self.waiting_queue[session_id]
            self.never_charged += 1
        elif station_id in self._EVSEs:
            if session_id is None:
                raise ValueError(
                    "StochasticNetwork requires a session_id to unplug an EV."
                )
            elif self._EVSEs[station_id].ev is None:
                pass
                # warnings.warn(
                #     f"Tried to remove EV with session_id {session_id} which was not "
                #     f"present at station {station_id}. Found no EV instead."
                # )
            elif session_id == self._EVSEs[station_id].ev.session_id:
                self._EVSEs[station_id].unplug()
                if len(self.waiting_queue) > 0:
                    _, next_ev = self.waiting_queue.popitem(last=False)
                    next_ev.update_station_id(station_id)
                    super().plugin(next_ev)
                    self.swaps += 1
        else:
            raise KeyError("Station {0} not found.".format(station_id))

    def post_charging_update(self):
        """ Unplug fully charged EVs, even if they are not scheduled to depart. """
        if self.early_departure:
            fully_charged_evs = [
                evse.ev
                for evse in self._EVSEs.values()
                if evse.ev is not None and evse.ev.fully_charged
            ]
            for ev in fully_charged_evs:
                if len(self.waiting_queue) > 0:
                    self.unplug(ev.station_id, ev.session_id)
                    self.early_unplug += 1



from unittest import TestCase
from unittest.mock import Mock, create_autospec

from acnportal.acnsim.models import Battery
from acnportal.acnsim.models import EV


class TestEV(TestCase):
    def setUp(self):
        basicBatt = create_autospec(Battery)
        self.ev = EV(0, 10, 25.0, "PS-001", "0001", basicBatt)

    def test_charge_valid_rate(self):
        self.ev._battery.charge = Mock(return_value=16)
        rate = self.ev.charge(16, 240, 5)
        self.assertEqual(rate, 16)
        self.assertAlmostEqual(self.ev.energy_delivered, 0.32)
        self.ev._battery.charge.assert_called_once()

    def test_charge_over_max_rate(self):
        self.ev._battery.charge = Mock(return_value=32)
        rate = self.ev.charge(40, 240, 5)
        self.assertEqual(rate, 32)
        self.assertAlmostEqual(self.ev.energy_delivered, 0.64)
        self.ev._battery.charge.assert_called_once()

    def test_reset(self):
        self.ev.reset()
        self.assertEqual(self.ev.energy_delivered, 0)
        self.ev._battery.reset.assert_called_once()



from .. import ChargingNetwork
from ..current import Current
from ...models.evse import get_evse_by_type


def simple_acn(
    station_ids,
    evse_type="BASIC",
    voltage=208,
    aggregate_cap=150,
    network_type=ChargingNetwork,
):
    """ Create a simple, single-phase network with a single aggregate constraint.

    Args:
        station_ids (List[str]): List of station_ids to include in network.
        evse_type (str): Identifier for the EVSE to use. See get_evse_by_type.
        voltage (float): Default voltage at the EVSEs. [V]
        aggregate_cap (float): Limit aggregate power draw to aggregate_cap. Default: 150. [kW]
        network_type (ChargingNetwork like): Type to be returned. Should be ChargingNetwork or a subclass.

    Returns:
        ChargingNetwork: A ChargingNetwork-like object configured with the EVSEs and constraints of the Caltech ACN.
    """
    network = network_type()
    for evse_id in station_ids:
        network.register_evse(get_evse_by_type(evse_id, evse_type), voltage, 0)
    agg = Current(station_ids)

    # Build constraint set
    current_cap = (aggregate_cap / voltage) * 1000
    network.add_constraint(agg, current_cap, name="Aggregate Current")

    return network



# coding=utf-8
""" Simple algorithm for uncontrolled EV charging. """
from typing import List

from .base_algorithm import BaseAlgorithm
from ..acnsim.interface import SessionInfo


class UncontrolledCharging(BaseAlgorithm):
    """ Simple algorithm for uncontrolled EV charging.

    Implements abstract class BaseAlgorithm.

    This algorithm ignores all infrastructure constraints.
    All EVs will be charged as quickly as possible according to their maximum charging
    rate and their battery dynamics.
    """

    def __init__(self) -> None:
        super().__init__()
        # Call algorithm each period since it only returns a rate for the next period.
        self.max_recompute = 1

    def schedule(self, active_sessions: List[SessionInfo]):
        """ Schedule each EV to charge as quickly as possible ignoring all
        infrastructure constraints.

        Implements abstract method schedule from BaseAlgorithm.

        See class documentation for description of the algorithm.

        This algorithm SHOULD NOT be run on physical infrastructure which has safety
        constraints as it could (and likely will) violate those constraints.

        Args:
            active_sessions (List[SessionInfo]): see BaseAlgorithm

        Returns:
            Dict[str, List[float]]: see BaseAlgorithm
        """
        schedule = {}
        for session in active_sessions:
            schedule[session.station_id] = [
                self.interface.max_pilot_signal(session.station_id)
            ]
        return schedule



import unittest
from unittest import TestCase
from unittest.mock import patch

from acnportal.acnsim.models import Battery, Linear2StageBattery, batt_cap_fn


class TestBatteryBase(TestCase):
    def setUp(self):
        self.init_charge = 50
        self.batt = Battery(100, self.init_charge, 32)

    def test_charge_negative_voltage(self):
        with self.assertRaises(ValueError):
            self.batt.charge(32, -240, 5)

    def test_charge_negative_period(self):
        with self.assertRaises(ValueError):
            self.batt.charge(32, 240, -5)

    def test_valid_reset_default(self):
        self.batt.charge(16, 240, 5)
        self.batt.reset()
        self.assertEqual(self.batt.current_charging_power, 0)
        self.assertEqual(self.batt._current_charge, self.init_charge)

    def test_valid_reset(self):
        self.batt.charge(16, 240, 5)
        self.batt.reset(45)
        self.assertEqual(self.batt.current_charging_power, 0)
        self.assertEqual(self.batt._current_charge, 45)

    def test_reset_over_capacity(self):
        with self.assertRaises(ValueError):
            self.batt.reset(110)


class TestBattery(TestBatteryBase):
    def setUp(self):
        self.init_charge = 50
        self.batt = Battery(100, self.init_charge, 7.68)

    def test_valid_charge(self):
        rate = self.batt.charge(16, 240, 5)
        self.assertEqual(rate, 16)
        self.assertEqual(self.batt.current_charging_power, 3.84)
        self.assertAlmostEqual(self.batt._current_charge, 50.32)

    def test_charge_over_max_rate(self):
        rate = self.batt.charge(55, 240, 5)
        self.assertEqual(rate, 32)
        self.assertEqual(self.batt.current_charging_power, 7.68)
        self.assertAlmostEqual(self.batt._current_charge, 50.64)

    def test_charge_over_capacity(self):
        self.batt = Battery(50, 49.8, 7.68)
        rate = self.batt.charge(32, 240, 5)
        self.assertAlmostEqual(rate, 10)
        self.assertAlmostEqual(self.batt.current_charging_power, 2.4)
        self.assertAlmostEqual(self.batt._current_charge, 50)


class TestLinear2StageBatteryStepwiseCharge(TestBatteryBase):
    def setUp(self):
        self.init_charge = 0
        self.batt = Linear2StageBattery(
            100, self.init_charge, 7.68, charge_calculation="stepwise"
        )

    def test_invalid_charge_calculation_method(self):
        with self.assertRaises(ValueError):
            self.batt = Linear2StageBattery(
                100, 0, 7.6, charge_calculation="invalid_method"
            )

    def test_negative_transition_soc(self):
        with self.assertRaises(ValueError):
            self.batt = Linear2StageBattery(
                100, 0, 7.68, transition_soc=-0.1, charge_calculation="stepwise"
            )

    def test_one_transition_soc(self):
        with self.assertRaises(ValueError):
            self.batt = Linear2StageBattery(
                100, 0, 7.68, transition_soc=1, charge_calculation="stepwise"
            )

    def test_over_one_transitions_soc(self):
        with self.assertRaises(ValueError):
            self.batt = Linear2StageBattery(
                100, 0, 7.68, transition_soc=1.1, charge_calculation="stepwise"
            )

    def test_zero_pilot_charge(self):
        self.batt = Linear2StageBattery(100, 0, 7.68, charge_calculation="stepwise")
        with patch("numpy.random.normal", return_value=1.2):
            rate = self.batt.charge(0, 240, 5)
        self.assertAlmostEqual(rate, 0)
        self.assertAlmostEqual(self.batt.current_charging_power, 0)
        self.assertAlmostEqual(self.batt._current_charge, 0)

    def test_valid_charge_no_noise_not_tail(self):
        self.batt = Linear2StageBattery(100, 0, 7.68, charge_calculation="stepwise")
        with patch("numpy.random.normal", return_value=1.2):
            rate = self.batt.charge(16, 240, 5)
        self.assertAlmostEqual(rate, 16)
        self.assertAlmostEqual(self.batt.current_charging_power, 3.84)
        self.assertAlmostEqual(self.batt._current_charge, 0.32)

    def test_valid_charge_positive_noise_not_tail(self):
        self.batt = Linear2StageBattery(
            100, 0, 7.68, noise_level=1, charge_calculation="stepwise"
        )
        with patch("numpy.random.normal", return_value=0.288):
            rate = self.batt.charge(16, 240, 5)
        self.assertAlmostEqual(rate, 14.8)
        self.assertAlmostEqual(self.batt.current_charging_power, 3.552)
        self.assertAlmostEqual(self.batt._current_charge, 0.296)

    def test_valid_charge_negative_noise_not_tail(self):
        self.batt = Linear2StageBattery(
            100, 0, 7.68, noise_level=1, charge_calculation="stepwise"
        )
        with patch("numpy.random.normal", return_value=-0.288):
            rate = self.batt.charge(16, 240, 5)
        self.assertAlmostEqual(rate, 14.8)
        self.assertAlmostEqual(self.batt.current_charging_power, 3.552)
        self.assertAlmostEqual(self.batt._current_charge, 0.296)

    def test_valid_charge_no_noise_tail(self):
        self.batt = Linear2StageBattery(100, 85, 7.68, charge_calculation="stepwise")
        rate = self.batt.charge(32, 240, 5)
        self.assertAlmostEqual(rate, 24)
        self.assertAlmostEqual(self.batt.current_charging_power, 5.76)
        self.assertAlmostEqual(self.batt._current_charge, 85.48)

    def test_valid_charge_positive_noise_tail(self):
        self.batt = Linear2StageBattery(
            100, 85, 7.68, noise_level=1, charge_calculation="stepwise"
        )
        with patch("numpy.random.normal", return_value=0.288):
            rate = self.batt.charge(32, 240, 5)
        self.assertAlmostEqual(rate, 25.2)
        self.assertAlmostEqual(self.batt.current_charging_power, 6.048)
        self.assertAlmostEqual(self.batt._current_charge, 85.504)

    def test_valid_charge_negative_noise_tail(self):
        self.batt = Linear2StageBattery(
            100, 85, 7.68, noise_level=1, charge_calculation="stepwise"
        )
        with patch("numpy.random.normal", return_value=-0.288):
            rate = self.batt.charge(32, 240, 5)
        self.assertAlmostEqual(rate, 22.8)
        self.assertAlmostEqual(self.batt.current_charging_power, 5.472)
        self.assertAlmostEqual(self.batt._current_charge, 85.456)

    def test_charge_over_max_rate_not_tail(self):
        self.batt = Linear2StageBattery(100, 0, 7.68, charge_calculation="stepwise")
        rate = self.batt.charge(40, 240, 5)
        self.assertAlmostEqual(rate, 32)
        self.assertAlmostEqual(self.batt.current_charging_power, 7.68)
        self.assertAlmostEqual(self.batt._current_charge, 0.64)

    def test_charge_over_capacity(self):
        self.batt = Linear2StageBattery(100, 99, 7.68, charge_calculation="stepwise")
        rate = self.batt.charge(32, 240, 5)
        self.assertAlmostEqual(rate, 1.6)
        self.assertAlmostEqual(self.batt.current_charging_power, 0.384)
        self.assertAlmostEqual(self.batt._current_charge, 99.032)


class TestLinear2StageBattery(TestLinear2StageBatteryStepwiseCharge):
    def setUp(self):
        self.init_charge = 0
        self.batt = Linear2StageBattery(100, self.init_charge, 7.68)

    def test_valid_charge_no_noise_tail(self):
        self.batt = Linear2StageBattery(100, 85, 7.68)
        rate = self.batt.charge(32, 240, 5)
        self.assertAlmostEqual(rate, 23.62006344060197)
        self.assertAlmostEqual(self.batt.current_charging_power, 5.668815225744472)
        self.assertAlmostEqual(self.batt._current_charge, 85.472401268812)

    def test_valid_charge_positive_noise_tail(self):
        self.batt = Linear2StageBattery(100, 85, 7.68, noise_level=1)
        with patch("numpy.random.normal", return_value=0.288):
            rate = self.batt.charge(32, 240, 5)
        self.assertAlmostEqual(rate, 22.42006344060197)
        self.assertAlmostEqual(self.batt.current_charging_power, 5.380815225744472)
        self.assertAlmostEqual(self.batt._current_charge, 85.448401268812)

    def test_valid_charge_negative_noise_tail(self):
        self.batt = Linear2StageBattery(100, 85, 7.68, noise_level=1)
        with patch("numpy.random.normal", return_value=-0.288):
            rate = self.batt.charge(32, 240, 5)
        self.assertAlmostEqual(rate, 22.42006344060197)
        self.assertAlmostEqual(self.batt.current_charging_power, 5.380815225744472)
        self.assertAlmostEqual(self.batt._current_charge, 85.448401268812)

    def test_charge_over_max_rate_not_tail(self):
        self.batt = Linear2StageBattery(100, 0, 7.68)
        rate = self.batt.charge(40, 240, 5)
        self.assertAlmostEqual(rate, 32)
        self.assertAlmostEqual(self.batt.current_charging_power, 7.68)
        self.assertAlmostEqual(self.batt._current_charge, 0.64)

    def test_charge_over_capacity(self):
        self.batt = Linear2StageBattery(100, 99, 7.68)
        rate = self.batt.charge(32, 240, 5)
        self.assertAlmostEqual(rate, 1.574670896040131)
        self.assertAlmostEqual(self.batt.current_charging_power, 0.3779210150496315)
        self.assertAlmostEqual(self.batt._current_charge, 99.0314934179208)

    def test_charge_at_threshold(self):
        self.batt = Linear2StageBattery(100, 80, 7.68)
        rate = self.batt.charge(32, 240, 5)
        self.assertAlmostEqual(rate, 31.49341792080207)
        self.assertAlmostEqual(self.batt.current_charging_power, 7.558420300992497)
        self.assertAlmostEqual(self.batt._current_charge, 80.629868358416)

    def test_charge_cross_threshold(self):
        self.batt = Linear2StageBattery(100, 79.9, 7.68)
        rate = self.batt.charge(32, 240, 5)
        self.assertAlmostEqual(rate, 31.63875847566334)
        self.assertAlmostEqual(self.batt.current_charging_power, 7.5933020341592)
        self.assertAlmostEqual(self.batt._current_charge, 80.5327751695133)


class TestBatteryFit(TestCase):
    def battery_feasible(self, request, duration, voltage, period):
        cap, init = batt_cap_fn(request, duration, voltage, period)
        batt = Linear2StageBattery(cap, init, 32 * voltage / 1000)
        rates = [batt.charge(32, voltage, period) for _ in range(duration)]
        self.assertAlmostEqual((batt._current_charge - init), request)
        self.assertAlmostEqual(request, sum(rates) * voltage / 1000 * (period / 60))

    def test_no_laxity(self):
        voltage = 208
        period = 5
        max_power = 32 * voltage / 1000
        for dur in [12, 24, 32, 64]:
            self.battery_feasible(max_power * dur / (60 / period), dur, voltage, period)

    def test_half_laxity(self):
        voltage = 208
        period = 5
        max_power = 32 * voltage / 1000
        for dur in [12, 24, 32, 64]:
            self.battery_feasible(
                max_power * dur / (60 / period) / 2, dur, voltage, period
            )

    def test_almost_no_laxity(self):
        voltage = 208
        period = 5
        max_power = 32 * voltage / 1000
        for dur in [12, 24, 32, 64]:
            self.battery_feasible(
                max_power * dur / (60 / period) / 1.001, dur, voltage, period
            )


if __name__ == "__main__":
    unittest.main()



import warnings
from typing import Optional, Dict, Any, Tuple

import numpy as np
from ..base import BaseSimObj

BASIC = "BASIC"
AV = "AeroVironment"
CC = "ClipperCreek"


def get_evse_by_type(station_id, evse_type):
    """ Factory to produce EVSEs of a given type.

    Args:
        station_id (str): Unique identifier of the EVSE.
        evse_type (str): Type of the EVSE. Currently supports 'BASIC', 'AeroVironment', and 'ClipperCreek'.

    Returns:
        EVSE: an EVSE of the specified type and with the specified id.
    """
    if evse_type == BASIC:
        return EVSE(station_id, max_rate=32)
    elif evse_type == AV:
        allowable_rates = [0]
        allowable_rates.extend(i for i in range(6, 33))
        return FiniteRatesEVSE(station_id, allowable_rates)
    elif evse_type == CC:
        allowable_rates = [0, 8, 16, 24, 32]
        return FiniteRatesEVSE(station_id, allowable_rates)


class InvalidRateError(Exception):
    """ Raised when an invalid pilot signal is passed to an EVSE. """

    pass


class StationOccupiedError(Exception):
    """ Raised when a plugin event is called for an EVSE that already has an EV attached. """


class BaseEVSE(BaseSimObj):
    """ Abstract base class to model Electric Vehicle Supply Equipment
    (charging station). This class is meant to be inherited from to
    implement new EVSEs.

    Subclasses must implement the max_rate, allowable_pilot_signals,
    and _valid_rate methods.

    Attributes:
        _station_id (str): Unique identifier of the EVSE.
        _ev (EV): EV currently connected the the EVSE.
        _current_pilot (float): Pilot signal for the current time step.
            [acnsim units]
        is_continuous (bool): If True, this EVSE accepts a continuous
            range of pilot signals. If False, this EVSE accepts only
            a discrete set of pilot signals.
    """

    def __init__(self, station_id):
        """ Initialize a BaseEVSE instance.

        Args:
            station_id (str): Unique identifier of the EVSE.
        """
        self._station_id = station_id
        self._ev = None
        self._current_pilot = 0
        self.is_continuous = True

    @property
    def station_id(self):
        """ Return unique identifier of the EVSE. (str) """
        return self._station_id

    @property
    def ev(self):
        """ Return EV currently connected the the EVSE. (EV) """
        return self._ev

    @property
    def max_rate(self):
        """ Return maximum charging current allowed by the EVSE. (float) """
        raise NotImplementedError

    @property
    def min_rate(self):
        """ Return minimum charging current allowed by the EVSE. (float) """
        return 0

    @property
    def current_pilot(self):
        """ Return pilot signal for the current time step. (float)"""
        return self._current_pilot

    @property
    def allowable_pilot_signals(self):
        """ Returns the allowable pilot signal levels for this EVSE.

        NOT IMPLEMENTED IN BaseEVSE. This method MUST be implemented in
        all subclasses.

        Returns:
            List[float]: List of acceptable pilot signal values or an
                interval of acceptable pilot signal values.
        """
        raise NotImplementedError

    def set_pilot(self, pilot, voltage, period):
        """ Apply a new pilot signal to the EVSE.

        Before applying the new pilot, this method first checks if the pilot is allowed. If it is not, an
        InvalidRateError is raised. If the rate is valid, it is forwarded on to the attached EV if one is present.
        This method is also where EV charging is triggered. Thus it must be called in every time time period where the
        attached EV should receive charge.

        Args:
            pilot (float): New pilot (control signal) to be sent to the attached EV. [A]
            voltage (float): AC voltage provided to the battery charger. [V]
            period (float): Length of the charging period. [minutes]

        Returns:
            None.

        Raises:
            InvalidRateError: Exception raised when pilot is not allowed by the EVSE.
        """
        if self._valid_rate(pilot):
            self._current_pilot = pilot
            if self._ev is not None:
                self._ev.charge(pilot, voltage, period)
        else:
            raise InvalidRateError(
                f"Pilot {pilot} A is not valid for " f"station {self.station_id}."
            )

    def _valid_rate(self, pilot, atol=1e-3):
        """ Check if pilot is in the valid set.

        NOT IMPLEMENTED IN BaseEVSE. This method MUST be implemented in
        all subclasses.

        Args:
            pilot (float): Proposed pilot signal.
            atol: Absolute tolerance used when determining if a pilot
                belongs to the allowable rates set.

        Returns:
            bool: True if the proposed pilot signal is valid. False
                otherwise.
        """
        raise NotImplementedError

    def plugin(self, ev):
        """ Method to attach an EV to the EVSE.

        Args:
            ev (EV): EV which should be attached to the EVSE.

        Returns:
            None.

        Raises:
            StationOccupiedError: Exception raised when plugin is called by an EV is already attached to the EVSE.
        """
        # assert ev.station_id == self.station_id
        if self.ev is None:
            self._ev = ev
        else:
            raise StationOccupiedError(
                f"Station {self._station_id} is occupied with ev "
                f"{self._ev.session_id}."
            )

    def unplug(self):
        """ Method to remove an EV currently attached to the EVSE.

        Sets ev to None and current_pilot to 0.

        Returns:
            None
        """
        self._ev = None
        self._current_pilot = 0

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """ Implements BaseSimObj._to_dict. """
        attribute_dict = {}
        nn_attr_lst = ["_station_id", "_current_pilot", "is_continuous"]
        for attr in nn_attr_lst:
            attribute_dict[attr] = getattr(self, attr)

        if self._ev is not None:
            # noinspection PyProtectedMember
            registry, context_dict = self.ev._to_registry(context_dict=context_dict)
            attribute_dict["_ev"] = registry["id"]
        else:
            attribute_dict["_ev"] = None

        return attribute_dict, context_dict

    @classmethod
    def _from_dict_helper(cls, out_obj, attribute_dict, context_dict, loaded_dict):
        out_obj._current_pilot = attribute_dict["_current_pilot"]
        out_obj.is_continuous = attribute_dict["is_continuous"]

        if attribute_dict["_ev"] is not None:
            # noinspection PyProtectedMember
            ev, loaded_dict = BaseSimObj._build_from_id(
                attribute_dict["_ev"], context_dict, loaded_dict=loaded_dict
            )
        else:
            ev = None
        out_obj._ev = ev
        return out_obj, loaded_dict

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """ Implements BaseSimObj._from_dict. """
        out_obj = cls(attribute_dict["_station_id"])
        return cls._from_dict_helper(out_obj, attribute_dict, context_dict, loaded_dict)


class EVSE(BaseEVSE):
    # TODO: For this class, is a 0 pilot assumed to be allowed?
    """ This class of EVSE allows for charging in a continuous range
    from min_rate to max_rate.

    Attributes:
        See BaseEVSE attributes.
        _max_rate (float): Maximum charging current allowed by the EVSE.
        _min_rate (float): Minimum charging current allowed by the EVSE.
    """

    def __init__(self, station_id, max_rate=float("inf"), min_rate=0):
        """ Initialize an EVSE instance.

        Args:
            See BaseEVSE __init__() Args.
            max_rate (float): Maximum charging current allowed by the
                EVSE.
            min_rate (float): Minimum charging current allowed by the
                EVSE.
        """
        super().__init__(station_id)
        self._max_rate = max_rate
        self._min_rate = min_rate

    @property
    def max_rate(self):
        """ Return maximum charging current allowed by the EVSE. (float) """
        return self._max_rate

    @property
    def min_rate(self):
        """ Return minimum charging current allowed by the EVSE. (float) """
        return self._min_rate

    @property
    def allowable_pilot_signals(self):
        """ Returns the allowable pilot signal levels for this EVSE.

        Implements abstract method allowable_pilot_signals from
        BaseEVSE.

        Returns:
            list[float]: List of 2 values: the min and max
                acceptable values.
        """
        return [self.min_rate, self.max_rate]

    def _valid_rate(self, pilot, atol=1e-3):
        """ Check if pilot is in the valid set.

        Implements abstract method _valid_rate from BaseEVSE.

        Args:
            pilot (float): Proposed pilot signal.
            atol: Absolute tolerance used when determining if a pilot
                belongs to the allowable rates set.

        Returns:
            bool: True if the proposed pilot signal is valid. False
                otherwise.
        """
        return self.min_rate <= pilot + atol and pilot - atol <= self.max_rate

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """ Implements BaseSimObj._to_dict. """
        attribute_dict, context_dict = super()._to_dict(context_dict)
        attribute_dict["_max_rate"] = self._max_rate
        attribute_dict["_min_rate"] = self._min_rate

        return attribute_dict, context_dict

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """ Implements BaseSimObj._from_dict. """
        out_obj = cls(
            attribute_dict["_station_id"],
            max_rate=attribute_dict["_max_rate"],
            min_rate=attribute_dict["_min_rate"],
        )
        return cls._from_dict_helper(out_obj, attribute_dict, context_dict, loaded_dict)


class DeadbandEVSE(BaseEVSE):
    """ Subclass of BaseEVSE which enforces the J1772 deadband between
    0 - 6 A.

    Attributes:
        See BaseEVSE attributes.
        _max_rate (float): Maximum charging current allowed by the EVSE.
        _deadband_end (float): Upper end of the deadband. Pilot signals
            between 0 and this number are not allowed for this EVSE.

    """

    def __init__(
        self, station_id, deadband_end=6, max_rate=float("inf"), min_rate=None
    ):
        """ Initialize a DeadbandEVSE instance.

        Args:
            See EVSE __init__() Args.
            max_rate (float): Maximum charging current allowed by the
                EVSE.
            deadband_end (float): Upper end of the deadband. Pilot
                signals between 0 and this number are not allowed for
                this EVSE.
        """
        super().__init__(station_id)
        self._max_rate = max_rate
        self._deadband_end = deadband_end
        if min_rate is not None:
            warnings.warn(
                f"Keyword argument 'min_rate' is deprecated for class "
                f"DeadbandEVSE. Providing 'min_rate' will raise an "
                f"error in a future release of acnportal.",
                DeprecationWarning,
            )

    @property
    def max_rate(self):
        """ Return maximum charging current allowed by the EVSE. (float) """
        return self._max_rate

    @property
    def deadband_end(self):
        """ Return deadband end of the EVSE. (float) """
        return self._deadband_end

    @property
    def allowable_pilot_signals(self):
        """ Returns the allowable pilot signal levels for this EVSE.

        It is implied that a 0 A signal is allowed.

        Implements abstract method allowable_pilot_signals from
        BaseEVSE.

        Returns:
            list[float]: List of 2 values: the min and max
                acceptable values.
        """
        return [self._deadband_end, self.max_rate]

    def _valid_rate(self, pilot, atol=1e-3):
        """ Check if pilot is in the valid set.

        Overrides super class method. Disallows rates between
        0 - 6 A as per the J1772 standard.

        Args:
            pilot: Proposed pilot signal.
            atol: Absolute tolerance used when determining if a pilot
                belongs to the allowable rates set.

        Returns:
            bool: True if the proposed pilot signal is valid. False
                otherwise.
        """
        return np.isclose(pilot, 0, atol=atol, rtol=0) or (
            self._deadband_end <= pilot + atol and pilot - atol <= self.max_rate
        )

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """ Implements BaseSimObj._to_dict. """
        attribute_dict, context_dict = super()._to_dict(context_dict)
        attribute_dict["_max_rate"] = self._max_rate
        attribute_dict["_deadband_end"] = self._deadband_end
        return attribute_dict, context_dict

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """ Implements BaseSimObj._from_dict. """
        out_obj = cls(
            attribute_dict["_station_id"],
            deadband_end=attribute_dict["_deadband_end"],
            max_rate=attribute_dict["_max_rate"],
        )
        return cls._from_dict_helper(out_obj, attribute_dict, context_dict, loaded_dict)


class FiniteRatesEVSE(BaseEVSE):
    """ Subclass of EVSE which allows for finite allowed rate sets.

    Most functionality remains the same except those differences noted
    below.

    Attributes:
        See BaseEVSE attributes.
        allowable_rates (iterable): Iterable of rates which are allowed
            by the EVSE. On initialization, allowable_rates is converted
            into a list of rates in increasing order that includes 0 and
            contains no duplicate values.

    """

    def __init__(self, station_id, allowable_rates):
        """ Initialize a DeadbandEVSE instance.

        Args:
            See EVSE __init__() Args.
            allowable_rates (iterable): Iterable of rates which are
                allowed by the EVSE. On initialization, allowable_rates
                is converted into a list of rates in increasing order
                that includes 0 and contains no duplicate values.
        """
        super().__init__(station_id)
        allowable_rates = set(allowable_rates)
        allowable_rates.add(0)
        self.allowable_rates = sorted(list(allowable_rates))
        self.is_continuous = False

    @property
    def max_rate(self):
        """ Return maximum charging current allowed by the EVSE. (float) """
        return max(self.allowable_rates)

    @property
    def min_rate(self):
        """ Return minimum charging current allowed by the EVSE. (float) """
        allowable_gt_zero = [r for r in self.allowable_rates if r > 0]
        if len(allowable_gt_zero) > 0:
            return min(allowable_gt_zero)
        else:
            return 0

    @property
    def allowable_pilot_signals(self):
        """ Returns the allowable pilot signal levels for this EVSE.

        Implements abstract method allowable_pilot_signals from
        BaseEVSE.

        Returns:
            list[float]: List of allowable pilot signals.
        """
        return self.allowable_rates

    def _valid_rate(self, pilot, atol=1e-3):
        """ Check if pilot is in the valid set.

        Overrides super class method. Checks if pilot is close to being
        in the allowable set.

        Args:
            pilot: Proposed pilot signal.
            atol: Absolute tolerance used when determining if a pilot
                belongs to the allowable rates set.

        Returns:
            bool: True if the proposed pilot signal is valid. False
                otherwise.
        """
        return np.any(np.isclose(pilot, self.allowable_rates, atol=1e-3, rtol=0))

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """ Implements BaseSimObj._to_dict. """
        attribute_dict, context_dict = super()._to_dict(context_dict)
        attribute_dict["allowable_rates"] = self.allowable_rates
        return attribute_dict, context_dict

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """ Implements BaseSimObj._from_dict. """
        out_obj = cls(attribute_dict["_station_id"], attribute_dict["allowable_rates"])
        return cls._from_dict_helper(out_obj, attribute_dict, context_dict, loaded_dict)



# coding=utf-8
""" Tests for StochasticChargingNetwork functionality. """
from unittest import TestCase
from unittest.mock import Mock, patch

from acnportal.contrib.acnsim.network import StochasticNetwork
from acnportal.acnsim.models import EV, EVSE, Battery

RANDOM_CHOICE_PATCH_STR = "acnportal.contrib.acnsim.stochastic_network.random.choice"


class TestStochasticChargingNetwork(TestCase):
    def setUp(self) -> None:
        """ Set up tests for ChargingNetwork. """
        self.network = StochasticNetwork()
        self.evse_ids = ["PS-001", "PS-002", "PS-003"]
        for evse_id in self.evse_ids:
            self.network.register_evse(EVSE(evse_id), 240, 0)

    def test_available_evses(self) -> None:
        result = self.network.available_evses()
        self.assertListEqual(result, self.evse_ids)

    @patch(RANDOM_CHOICE_PATCH_STR)
    def test_plugin_ev_with_spot_available(self, choice_mock) -> None:
        ev = EV(0, 10, 10, "", "Session-01", Battery(100, 0, 6.6))
        choice_mock.return_value = "PS-002"
        self.network.plugin(ev)
        self.assertEqual(ev.station_id, "PS-002")
        self.assertIs(self.network.get_ev("PS-002"), ev)

    def test_plugin_ev_no_spot_available_empty_waiting_queue(self) -> None:
        ev = EV(0, 10, 10, "", "Session-01", Battery(100, 0, 6.6))

        # Mock available_evses to behavior like the network is full.
        self.network.available_evses = Mock(return_value=[])

        self.network.plugin(ev)

        # EV should be added to waiting_queue
        self.assertEqual(len(self.network.waiting_queue), 1)
        self.assertIn(ev.session_id, self.network.waiting_queue)
        self.assertIs(self.network.waiting_queue[ev.session_id], ev)

    def test_plugin_ev_no_spot_available_evs_in_waiting_queue(self) -> None:
        ev1 = EV(0, 10, 10, "", "Session-01", Battery(100, 0, 6.6))
        ev2 = EV(0, 10, 10, "", "Session-02", Battery(100, 0, 6.6))

        # Mock available_evses to behavior like the network is full.
        self.network.available_evses = Mock(return_value=[])

        # Add ev1 to the waiting queue
        self.network.waiting_queue[ev1.session_id] = ev1
        self.network.waiting_queue.move_to_end(ev1.session_id)

        self.network.plugin(ev2)

        # EV should be added to waiting_queue
        self.assertEqual(len(self.network.waiting_queue), 2)

        # ev1 should be after ev2 in the queue
        _, first_in_queue = self.network.waiting_queue.popitem(last=False)
        _, second_in_queue = self.network.waiting_queue.popitem(last=False)
        self.assertIs(first_in_queue, ev1)
        self.assertIs(second_in_queue, ev2)

    @patch(RANDOM_CHOICE_PATCH_STR)
    def test_unplug_station_exists_session_id_matches_queue_empty(
        self, choice_mock
    ) -> None:
        ev = EV(0, 10, 10, "", "Session-01", Battery(100, 0, 6.6))
        choice_mock.return_value = "PS-002"

        self.network.plugin(ev)

        self.network._EVSEs["PS-002"].unplug = Mock()
        self.network.unplug(ev.station_id, ev.session_id)
        self.network._EVSEs["PS-002"].unplug.assert_called()

    @patch(RANDOM_CHOICE_PATCH_STR)
    def test_unplug_station_exists_session_id_matches_ev_in_queue(
        self, choice_mock
    ) -> None:
        ev1 = EV(0, 10, 10, "", "Session-01", Battery(100, 0, 6.6))
        ev2 = EV(0, 10, 10, "", "Session-02", Battery(100, 0, 6.6))

        # Plug ev1 into space PS-001
        choice_mock.return_value = "PS-002"
        self.network.plugin(ev1)

        # Add ev2 to the waiting queue
        self.network.waiting_queue[ev2.session_id] = ev2
        self.network.waiting_queue.move_to_end(ev2.session_id)

        # Mock the plugin method so we can assert it was called.
        self.network._EVSEs["PS-002"].plugin = Mock()

        # Unplug ev1
        self.network.unplug(ev1.station_id, ev1.session_id)

        # ev2 should be plugged in after ev1 leaves
        self.network._EVSEs["PS-002"].plugin.assert_called_with(ev2)

        # swaps counter should be incremented
        self.assertEqual(self.network.swaps, 1)

    def test_unplug_when_ev_in_waiting_queue(self):
        ev1 = EV(0, 10, 10, "", "Session-01", Battery(100, 0, 6.6))

        # Add ev2 to the waiting queue
        self.network.waiting_queue[ev1.session_id] = ev1
        self.network.waiting_queue.move_to_end(ev1.session_id)

        self.assertEqual(len(self.network.waiting_queue), 1)

        self.network.unplug(ev1.station_id, ev1.session_id)

        # ev1 should be removed from the waiting queue
        self.assertEqual(len(self.network.waiting_queue), 0)

        # never_charged counter should be incremented
        self.assertEqual(self.network.never_charged, 1)

    @patch(RANDOM_CHOICE_PATCH_STR)
    def test_unplug_station_exists_session_id_mismatch(self, choice_mock) -> None:
        ev = EV(0, 10, 10, "", "Session-01", Battery(100, 0, 6.6))
        choice_mock.return_value = "PS-002"
        self.network.plugin(ev)
        self.network._EVSEs["PS-002"].unplug = Mock()
        self.network.unplug(ev.station_id, "Incorrect ID")
        # noinspection PyUnresolvedReferences
        self.network._EVSEs["PS-002"].unplug.assert_not_called()

    def test_unplug_station_does_not_exist(self) -> None:
        with self.assertRaises(KeyError):
            self.network.unplug("PS-005", "Session-01")

    @patch(RANDOM_CHOICE_PATCH_STR)
    def test_post_charging_update_early_departure_ev_in_queue(self, choice_mock):
        self.network.early_departure = True

        ev1 = EV(0, 10, 10, "", "Session-01", Battery(100, 0, 6.6))
        ev2 = EV(0, 10, 10, "", "Session-02", Battery(100, 0, 6.6))

        # Plug ev1 into space PS-001
        choice_mock.return_value = "PS-002"
        self.network.plugin(ev1)

        # Fully charge ev1
        ev1._energy_delivered = 10

        # Add ev2 to the waiting queue
        self.network.waiting_queue[ev2.session_id] = ev2
        self.network.waiting_queue.move_to_end(ev2.session_id)

        # Test that unplug is called during post_charging_update.
        self.network.unplug = Mock()
        self.network.post_charging_update()
        self.network.unplug.assert_called_with(ev1.station_id, ev1.session_id)
        self.assertEqual(self.network.early_unplug, 1)

    @patch(RANDOM_CHOICE_PATCH_STR)
    def test_post_charging_update_early_departure_off_ev_in_queue(self, choice_mock):
        self.network.early_departure = False

        ev1 = EV(0, 10, 10, "", "Session-01", Battery(100, 0, 6.6))
        ev2 = EV(0, 10, 10, "", "Session-02", Battery(100, 0, 6.6))

        # Plug ev1 into space PS-001
        choice_mock.return_value = "PS-002"
        self.network.plugin(ev1)

        # Fully charge ev1
        ev1._energy_delivered = 10

        # Add ev2 to the waiting queue
        self.network.waiting_queue[ev2.session_id] = ev2
        self.network.waiting_queue.move_to_end(ev2.session_id)

        # Test that unplug is called during post_charging_update.
        self.network.unplug = Mock()
        self.network.post_charging_update()
        self.network.unplug.assert_not_called()

    @patch(RANDOM_CHOICE_PATCH_STR)
    def test_post_charging_update_early_departure_ev_not_full(self, choice_mock):
        self.network.early_departure = True

        ev1 = EV(0, 10, 10, "", "Session-01", Battery(100, 0, 6.6))
        ev2 = EV(0, 10, 10, "", "Session-02", Battery(100, 0, 6.6))

        # Plug ev1 into space PS-001
        choice_mock.return_value = "PS-002"
        self.network.plugin(ev1)

        # ev1 not fully charged
        ev1._energy_delivered = 5

        # Add ev2 to the waiting queue
        self.network.waiting_queue[ev2.session_id] = ev2
        self.network.waiting_queue.move_to_end(ev2.session_id)

        # Test that unplug is called during post_charging_update.
        self.network.unplug = Mock()
        self.network.post_charging_update()
        self.network.unplug.assert_not_called()

    @patch(RANDOM_CHOICE_PATCH_STR)
    def test_post_charging_update_early_departure_no_ev_in_queue(self, choice_mock):
        self.network.early_departure = True

        ev1 = EV(0, 10, 10, "", "Session-01", Battery(100, 0, 6.6))

        # Plug ev1 into space PS-001
        choice_mock.return_value = "PS-002"
        self.network.plugin(ev1)

        # Fully charge ev1
        ev1._energy_delivered = 10

        # Test that unplug is called during post_charging_update.
        self.network.unplug = Mock()
        self.network.post_charging_update()
        self.network.unplug.assert_not_called()



from datetime import datetime
import pytz


def http_date(dt):
    """ Convert datetime object into http date according to RFC 1123.

    :param datetime dt: datetime object to convert
    :return: dt as a string according to RFC 1123 format
    :rtype: str
    """
    return dt.astimezone(pytz.utc).strftime("%a, %d %b %Y %H:%M:%S GMT")


def parse_http_date(ds, tz):
    """ Convert a string in RFC 1123 format to a datetime object

    :param str ds: string representing a datetime in RFC 1123 format.
    :param pytz.timezone tz: timezone to convert the string to as a pytz object.
    :return: datetime object.
    :rtype: datetime
    """
    dt = pytz.UTC.localize(datetime.strptime(ds, "%a, %d %b %Y %H:%M:%S GMT"))
    return dt.astimezone(tz)


def parse_dates(doc):
    """ Convert all datetime fields in RFC 1123 format in doc to datetime objects.

    :param dict doc: document to be converted as a dictionary.
    :return: doc with all RFC 1123 datetime fields replaced by datetime objects.
    :rtype: dict
    """
    tz = pytz.timezone(doc["timezone"])
    for field in doc:
        if isinstance(doc[field], str):
            try:
                dt = parse_http_date(doc[field], tz)
                doc[field] = dt
            except ValueError:
                pass
        if isinstance(doc[field], dict) and "timestamps" in doc[field]:
            doc[field]["timestamps"] = [
                parse_http_date(ds, tz) for ds in doc[field]["timestamps"]
            ]



# coding=utf-8
""" Tests for ChargingNetwork functionality. """
from collections import OrderedDict
from unittest import TestCase
from unittest.mock import Mock, create_autospec, patch

import numpy as np
import pandas as pd

from acnportal.acnsim import ChargingNetwork, Current
from acnportal.acnsim.models import EV, EVSE
from acnportal.acnsim.network.charging_network import EVSERegistrationError


class TestChargingNetwork(TestCase):
    def setUp(self) -> None:
        """ Set up tests for ChargingNetwork. """
        self.network = ChargingNetwork()
        self.network_with_tolerance = ChargingNetwork(
            violation_tolerance=1e-3, relative_tolerance=1e-5
        )

    def _test_init_empty(self, network: ChargingNetwork) -> None:
        self.assertEqual(network._EVSEs, OrderedDict())
        self.assertEqual(network.constraint_matrix, None)
        self.assertEqual(network.constraint_index, [])
        np.testing.assert_equal(network.magnitudes, np.array([]))
        np.testing.assert_equal(network._voltages, np.array([]))
        np.testing.assert_equal(network._phase_angles, np.array([]))

    def test_init_default_tolerance(self) -> None:
        self._test_init_empty(self.network)
        self.assertEqual(self.network.violation_tolerance, 1e-5)
        self.assertEqual(self.network.relative_tolerance, 1e-7)

    def test_init_default_tolerance_set(self) -> None:
        self._test_init_empty(self.network_with_tolerance)
        self.assertEqual(self.network_with_tolerance.violation_tolerance, 1e-3)
        self.assertEqual(self.network_with_tolerance.relative_tolerance, 1e-5)

    def test_register_evse_constraints_added_error(self) -> None:
        evse1 = EVSE("PS-001")
        self.network.register_evse(evse1, 240, -30)
        self.network.add_constraint(Current(["PS-001"]), 10)
        with self.assertRaises(EVSERegistrationError):
            self.network.register_evse(EVSE("PS-002"), 100, 150)

    def test_register_evse(self) -> None:
        evse1 = EVSE("PS-001")
        self.network.register_evse(evse1, 240, -30)
        evse2 = EVSE("PS-002")
        evse3 = EVSE("PS-003")
        self.network.register_evse(evse3, 100, 150)
        self.network.register_evse(evse2, 140, 90)
        self.assertIn("PS-001", self.network._EVSEs)
        self.assertIs(self.network._EVSEs["PS-001"], evse1)
        self.assertIn("PS-002", self.network._EVSEs)
        self.assertIs(self.network._EVSEs["PS-002"], evse2)
        self.assertIn("PS-003", self.network._EVSEs)
        self.assertIs(self.network._EVSEs["PS-003"], evse3)
        self.assertEqual(self.network.station_ids, ["PS-001", "PS-003", "PS-002"])
        np.testing.assert_allclose(self.network._phase_angles, np.array([-30, 150, 90]))
        np.testing.assert_allclose(self.network._voltages, np.array([240, 100, 140]))

    def test_plugin_station_exists(self) -> None:
        evse = EVSE("PS-001")
        evse.plugin = Mock(evse.plugin)
        self.network.register_evse(evse, 240, 0)
        ev = create_autospec(EV)
        ev.station_id = "PS-001"
        self.network.plugin(ev)
        # noinspection PyUnresolvedReferences
        evse.plugin.assert_called_once()

    def test_plugin_station_does_not_exist(self) -> None:
        ev = create_autospec(EV)
        ev.station_id = "PS-001"
        with self.assertRaises(KeyError):
            self.network.plugin(ev)

    def test_unplug_no_ev(self) -> None:
        evse = EVSE("PS-001")
        self.network.register_evse(evse, 240, 0)
        with patch.object(evse, "unplug") as unplug:
            with self.assertWarns(UserWarning):
                self.network.unplug("PS-001", "Session-01")
        unplug.assert_not_called()

    def _unplug_test_setup_helper(self, real_session_id: str) -> EVSE:
        evse = EVSE("PS-001")
        ev = create_autospec(EV)
        ev.station_id = "PS-001"
        ev.session_id = real_session_id
        evse.plugin(ev)

        self.network.register_evse(evse, 240, 0)
        return evse

    def test_unplug_no_session_id(self) -> None:
        evse = self._unplug_test_setup_helper("Session-01")
        with patch.object(evse, "unplug") as unplug:
            with self.assertWarns(UserWarning):
                self.network.unplug("PS-001")
        unplug.assert_called_once()

    def test_unplug_station_exists_session_id_matches(self) -> None:
        evse = self._unplug_test_setup_helper("Session-01")
        with patch.object(evse, "unplug") as unplug:
            self.network.unplug("PS-001", "Session-01")
        unplug.assert_called_once()

    def test_unplug_station_exists_session_id_mismatch(self) -> None:
        evse = self._unplug_test_setup_helper("Session-02")
        with patch.object(evse, "unplug") as unplug:
            with self.assertWarns(UserWarning):
                self.network.unplug("PS-001", "Session-01")
        unplug.assert_not_called()

    def test_unplug_station_does_not_exist(self) -> None:
        with self.assertRaises(KeyError):
            self.network.unplug("PS-001", "Session-01")

    def test_get_ev_station_exists(self) -> None:
        evse = EVSE("PS-001")
        ev = Mock(EV)
        evse.plugin(ev)
        self.network.register_evse(evse, 240, 0)
        self.assertIs(self.network.get_ev("PS-001"), ev)

    def test_get_ev_station_does_not_exist(self) -> None:
        with self.assertRaises(KeyError):
            self.network.get_ev("PS-001")

    def test_update_pilots_valid_pilots(self) -> None:
        evse1 = EVSE("PS-001")
        self.network.register_evse(evse1, 240, 0)
        evse2 = EVSE("PS-002")
        self.network.register_evse(evse2, 240, 0)
        evse3 = EVSE("PS-003")
        self.network.register_evse(evse3, 240, 0)
        evse1.set_pilot = create_autospec(evse1.set_pilot)
        evse2.set_pilot = create_autospec(evse2.set_pilot)
        evse3.set_pilot = create_autospec(evse3.set_pilot)
        self.network.update_pilots(np.array([[24, 16], [16, 24], [0, 0]]), 0, 5)
        # noinspection PyUnresolvedReferences
        evse1.set_pilot.assert_any_call(24, 240, 5)
        # noinspection PyUnresolvedReferences
        evse2.set_pilot.assert_any_call(16, 240, 5)
        # noinspection PyUnresolvedReferences
        evse3.set_pilot.assert_any_call(0, 240, 5)

    def _accessor_test_setup(self) -> None:
        evse1 = EVSE("PS-001")
        evse2 = EVSE("PS-002")
        evse3 = EVSE("PS-003")
        self.network.register_evse(evse1, 240, -30)
        self.network.register_evse(evse3, 100, 150)
        self.network.register_evse(evse2, 140, 90)

    def test_voltage_accessor(self) -> None:
        self._accessor_test_setup()
        self.assertEqual(
            self.network.voltages, {"PS-001": 240, "PS-003": 100, "PS-002": 140}
        )

    def test_phase_angles_accessor(self) -> None:
        self._accessor_test_setup()
        self.assertEqual(
            self.network.phase_angles, {"PS-001": -30, "PS-003": 150, "PS-002": 90}
        )


class TestChargingNetworkConstraints(TestCase):
    def setUp(self) -> None:
        """ Tests for constraint functionality of ChargingNetwork. """
        self.network = ChargingNetwork()
        self.network.register_evse(EVSE("PS-001"), 240, 0)
        self.network.register_evse(EVSE("PS-004"), 240, 0)
        self.network.register_evse(EVSE("PS-003"), 240, 0)
        self.network.register_evse(EVSE("PS-002"), 240, 0)
        self.network.register_evse(EVSE("PS-006"), 240, 0)
        curr_dict1 = {"PS-001": 0.25, "PS-002": 0.50, "PS-003": -0.25}
        current1 = Current(curr_dict1)
        curr_dict2 = {"PS-006": 0.30, "PS-004": -0.60, "PS-002": 0.50}
        current2 = Current(curr_dict2)
        self.network.add_constraint(current1, 50, name="first_constraint")
        self.network.add_constraint(current2, 10)

    def test_constraints_as_df(self) -> None:
        constraint_frame = self.network.constraints_as_df()
        pd.testing.assert_frame_equal(
            constraint_frame,
            pd.DataFrame(
                data=[[0.25, 0.00, -0.25, 0.50, 0.00], [0.00, -0.60, 0.00, 0.50, 0.30]],
                columns=["PS-001", "PS-004", "PS-003", "PS-002", "PS-006"],
                index=["first_constraint", "_const_1"],
            ),
        )

    def test_add_constraint_unregistered_evse(self) -> None:
        curr_dict3 = {"PS-001": -0.25, "PS-005": -0.50, "PS-003": 0.25}
        current3 = Current(curr_dict3)
        with self.assertRaises(KeyError):
            self.network.add_constraint(current3, 25, name="_bad_const")

    def test_add_constraint(self) -> None:
        np.testing.assert_allclose(self.network.magnitudes, np.array([50, 10]))
        np.testing.assert_allclose(
            self.network.constraint_matrix,
            np.array(
                [[0.25, 0.00, -0.25, 0.50, 0.00], [0.00, -0.60, 0.00, 0.50, 0.30]]
            ),
        )
        self.assertEqual(self.network.constraint_index[0], "first_constraint")
        self.assertEqual(self.network.constraint_index[1], "_const_1")

    def test_add_constraint_warning(self) -> None:
        curr_dict3 = {"PS-001": -0.25, "PS-002": -0.50, "PS-003": 0.25}
        current3 = Current(curr_dict3)
        with self.assertWarns(UserWarning):
            self.network.add_constraint(current3, 25, name="_const_1")
        np.testing.assert_allclose(self.network.magnitudes, np.array([50, 10, 25]))
        np.testing.assert_allclose(
            self.network.constraint_matrix,
            np.array(
                [
                    [0.25, 0.00, -0.25, 0.50, 0.00],
                    [0.00, -0.60, 0.00, 0.50, 0.30],
                    [-0.25, 0.00, 0.25, -0.50, 0.00],
                ]
            ),
        )
        self.assertEqual(self.network.constraint_index[0], "first_constraint")
        self.assertEqual(self.network.constraint_index[1], "_const_1")
        self.assertEqual(self.network.constraint_index[2], "_const_1_v2")

    def test_remove_constraint_not_found(self) -> None:
        with self.assertRaises(KeyError):
            self.network.remove_constraint("_const_100")

    def test_remove_constraint(self) -> None:
        self.network.remove_constraint("first_constraint")
        self.assertEqual(self.network.constraint_index, ["_const_1"])
        np.testing.assert_allclose(self.network.magnitudes, np.array([10]))
        np.testing.assert_allclose(
            self.network.constraint_matrix, np.array([[0.00, -0.60, 0.00, 0.50, 0.30]])
        )

    def test_update_constraint_not_found(self) -> None:
        with self.assertRaises(KeyError):
            self.network.update_constraint("_const_100", Current(), 0)

    def test_update_constraint(self) -> None:
        curr_dict3 = {"PS-001": -0.25, "PS-002": -0.50, "PS-003": 0.25}
        current3 = Current(curr_dict3)
        self.network.update_constraint("first_constraint", current3, 25)
        self.assertEqual(
            self.network.constraint_index, ["_const_1", "first_constraint"]
        )
        np.testing.assert_allclose(self.network.magnitudes, np.array([10, 25]))
        np.testing.assert_allclose(
            self.network.constraint_matrix,
            np.array(
                [[0.00, -0.60, 0.00, 0.50, 0.30], [-0.25, 0.00, 0.25, -0.50, 0.00]]
            ),
        )

    def test_update_constraint_new_name(self) -> None:
        curr_dict3 = {"PS-001": -0.25, "PS-002": -0.50, "PS-003": 0.25}
        current3 = Current(curr_dict3)
        self.network.update_constraint(
            "first_constraint", current3, 25, new_name="negated_first"
        )
        self.assertEqual(self.network.constraint_index, ["_const_1", "negated_first"])
        np.testing.assert_allclose(self.network.magnitudes, np.array([10, 25]))
        np.testing.assert_allclose(
            self.network.constraint_matrix,
            np.array(
                [[0.00, -0.60, 0.00, 0.50, 0.30], [-0.25, 0.00, 0.25, -0.50, 0.00]]
            ),
        )

    def test_is_feasible_good_loads(self) -> None:
        good_loads = np.array([[0, 0], [150, 120], [100, 40], [150, 120], [60, 9]])
        self.assertTrue(self.network.is_feasible(good_loads))

    def test_is_feasible_bad_loads(self) -> None:
        bad_loads = np.array([[0, 0], [150, 800], [100, 0], [150, 20], [60, 9]])
        self.assertFalse(self.network.is_feasible(bad_loads))

    def test_is_feasible_barely_good_loads(self) -> None:
        good_loads = np.array(
            [[160, 200.000035], [0, 0], [0, 0], [20.000015, 0], [0, 0]]
        )
        self.assertTrue(self.network.is_feasible(good_loads))

    def test_is_feasible_barely_bad_loads(self) -> None:
        bad_loads = np.array(
            [[160, 200.000045], [0, 0], [0, 0], [20.000015, 0], [0, 0]]
        )
        self.assertFalse(self.network.is_feasible(bad_loads))

    def test_is_feasible_barely_good_loads_set(self) -> None:
        good_loads = np.array([[160, 200.0035], [0, 0], [0, 0], [20.0015, 0], [0, 0]])
        self.assertTrue(self.network.is_feasible(good_loads, violation_tolerance=1e-3))

    def test_is_feasible_barely_bad_loads_set(self) -> None:
        bad_loads = np.array([[160, 200.0045], [0, 0], [0, 0], [20.0015, 0], [0, 0]])
        self.assertFalse(self.network.is_feasible(bad_loads, violation_tolerance=1e-3))

    def test_is_feasible_barely_good_loads_rel_set(self) -> None:
        good_loads = np.array([[160, 200.15], [0, 0], [0, 0], [20.015, 0], [0, 0]])
        self.assertTrue(self.network.is_feasible(good_loads, relative_tolerance=1e-3))

    def test_is_feasible_barely_bad_loads_rel_set(self) -> None:
        bad_loads = np.array([[160, 200.21], [0, 0], [0, 0], [20.021, 0], [0, 0]])
        self.assertFalse(self.network.is_feasible(bad_loads, relative_tolerance=1e-3))

    def test_is_feasible_barely_good_loads_set_linear(self) -> None:
        good_loads = np.array([[160, 200.0035], [0, 0], [0, 0], [20.0015, 0], [0, 0]])
        self.assertTrue(
            self.network.is_feasible(good_loads, violation_tolerance=1e-3, linear=True)
        )

    def test_is_feasible_barely_bad_loads_set_linear(self) -> None:
        bad_loads = np.array([[160, 200.0045], [0, 0], [0, 0], [20.0015, 0], [0, 0]])
        self.assertFalse(
            self.network.is_feasible(bad_loads, violation_tolerance=1e-3, linear=True)
        )

    def test_constraint_current(self) -> None:
        loads = np.array([[0, 0], [150, 120], [100, 40], [150, 120], [60, 9]])
        np.testing.assert_allclose(
            self.network.constraint_current(loads),
            np.array([[50 + 0j, 50 + 0j], [3 + 0j, -9.3 + 0j]]),
        )
        np.testing.assert_allclose(
            self.network.constraint_current(
                loads, constraints=["_const_1"], time_indices=[1]
            ),
            np.array([[-9.3 + 0j]]),
        )
        np.testing.assert_allclose(
            self.network.constraint_current(loads, constraints=["first_constraint"]),
            np.array([[50 + 0j, 50 + 0j]]),
        )
        np.testing.assert_allclose(
            self.network.constraint_current(loads, time_indices=[1]),
            np.array([[50 + 0j], [-9.3 + 0j]]),
        )



from builtins import property
from typing import Optional, Dict, Any, Tuple

from ..base import BaseSimObj


class EV(BaseSimObj):
    """Class to model the behavior of an Electrical Vehicle (ev).

    Args:
        arrival (int): Arrival time of the ev. [periods]
        departure (int): Departure time of the ev. [periods]
        requested_energy (float): Energy requested by the ev on arrival. [kWh]
        station_id (str): Identifier of the station used by this ev.
        session_id (str): Identifier of the session belonging to this ev.
        battery (Battery-like): Battery object to be used by the EV.
    """

    def __init__(
        self,
        arrival,
        departure,
        requested_energy,
        station_id,
        session_id,
        battery,
        estimated_departure=None,
    ):
        # User Defined Parameters
        self._arrival = arrival
        self._departure = departure
        self._session_id = session_id
        self._station_id = station_id

        # Estimate of session parameters
        self._requested_energy = requested_energy
        self._estimated_departure = (
            estimated_departure if estimated_departure is not None else departure
        )

        # Internal State
        self._battery = battery
        self._energy_delivered = 0
        self._current_charging_rate = 0

    @property
    def arrival(self):
        """ Return the arrival time of the EV."""
        return self._arrival

    @arrival.setter
    def arrival(self, value):
        """ Set the arrival time of the EV. (int) """
        self._arrival = value

    @property
    def departure(self):
        """ Return the departure time of the EV. (int) """
        return self._departure

    @departure.setter
    def departure(self, value):
        """ Set the departure time of the EV. (int) """
        self._departure = value

    @property
    def estimated_departure(self):
        """ Return the estimated departure time of the EV."""
        return self._estimated_departure

    @estimated_departure.setter
    def estimated_departure(self, value):
        """ Set the estimated departure time of the EV. (int) """
        self._estimated_departure = value

    @property
    def requested_energy(self):
        """ Return the energy request of the EV for this session. (float) [acnsim units]. """
        return self._requested_energy

    @property
    def session_id(self):
        """ Return the unique session identifier for this charging session. (str) """
        return self._session_id

    @property
    def station_id(self):
        """ Return the unique identifier for the EVSE used for this charging session. """
        return self._station_id

    @property
    def energy_delivered(self):
        """ Return the total energy delivered so far in this charging session. (float) """
        return self._energy_delivered

    @property
    def current_charging_rate(self):
        """ Return the current charging rate of the EV. (float) """
        return self._current_charging_rate

    @property
    def remaining_demand(self):
        """ Return the remaining energy demand of this session. (float)

        Defined as the difference between the requested energy of the session and the energy delivered so far.
        """
        return self.requested_energy - self.energy_delivered

    @property
    def fully_charged(self):
        """ Return True if the EV's demand has been fully met. (bool)"""
        return not (self.remaining_demand > 1e-3)

    @property
    def percent_remaining(self):
        """ Return the percent of demand which still needs to be fulfilled. (float)

        Defined as the ratio of remaining demand and requested energy. """
        return self.remaining_demand / self.requested_energy

    @property
    def maximum_charging_power(self):
        """ Return the maximum charging power of the battery."""
        return self._battery.max_charging_power

    def update_station_id(self, station_id):
        """ Method to update the station where EV will charge. """
        self._station_id = station_id

    def charge(self, pilot, voltage, period):
        """ Method to "charge" the ev.

        Args:
            pilot (float): Pilot signal passed to the battery. [A]
            voltage (float): AC voltage provided to the battery charger. [V]
            period (float): Length of the charging period. [minutes]

        Returns:
            float: Actual charging rate of the ev. [A]
        """
        charge_rate = self._battery.charge(pilot, voltage, period)
        self._energy_delivered += (charge_rate * voltage) / 1000 * (period / 60)
        self._current_charging_rate = charge_rate
        return charge_rate

    def reset(self):
        """ Reset battery back to its initialization. Also reset energy delivered.

        Returns:
            None.
        """
        self._energy_delivered = 0
        self._battery.reset()

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """ Implements BaseSimObj._to_dict. """
        attribute_dict = {}
        nn_attr_lst = [
            "_arrival",
            "_departure",
            "_session_id",
            "_station_id",
            "_requested_energy",
            "_estimated_departure",
            "_energy_delivered",
            "_current_charging_rate",
        ]
        for attr in nn_attr_lst:
            attribute_dict[attr] = getattr(self, attr)

        # noinspection PyProtectedMember
        registry, context_dict = self._battery._to_registry(context_dict=context_dict)
        attribute_dict["_battery"] = registry["id"]

        return attribute_dict, context_dict

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """ Implements BaseSimObj._from_dict. """
        # noinspection PyProtectedMember
        battery, loaded_dict = BaseSimObj._build_from_id(
            attribute_dict["_battery"], context_dict, loaded_dict=loaded_dict
        )

        out_obj = cls(
            attribute_dict["_arrival"],
            attribute_dict["_departure"],
            attribute_dict["_requested_energy"],
            attribute_dict["_station_id"],
            attribute_dict["_session_id"],
            battery,
            estimated_departure=attribute_dict["_estimated_departure"],
        )
        out_obj._energy_delivered = attribute_dict["_energy_delivered"]
        out_obj._current_charging_rate = attribute_dict["_current_charging_rate"]
        return out_obj, loaded_dict



# coding=utf-8
""" Tests for objects of the interface module. """
from typing import List
from unittest import TestCase
from unittest.mock import create_autospec, patch, Mock
import numpy.testing as nptest
import numpy as np

from acnportal.acnsim import (
    Simulator,
    Interface,
    InvalidScheduleError,
    FiniteRatesEVSE,
    DeadbandEVSE,
)
from acnportal.acnsim.interface import SessionInfo, InfrastructureInfo, Constraint
from acnportal.acnsim.models import EVSE
from acnportal.acnsim.network import ChargingNetwork


class TestSessionInfo(TestCase):
    def test_valid_inputs_w_defaults(self) -> None:
        s = SessionInfo("PS-001", "01", 10, 4, 5, 60)
        self.assertEqual(s.station_id, "PS-001")
        self.assertEqual(s.session_id, "01")
        self.assertEqual(s.requested_energy, 10)
        self.assertEqual(s.energy_delivered, 4)
        self.assertEqual(s.arrival, 5)
        self.assertEqual(s.departure, 60)
        self.assertEqual(s.estimated_departure, 60)
        self.assertEqual(s.current_time, 0)
        nptest.assert_array_equal(s.min_rates, 0)
        self.assertEqual(s.min_rates.shape, (55,))
        nptest.assert_array_equal(s.max_rates, float("inf"))
        self.assertEqual(s.max_rates.shape, (55,))
        self.assertEqual(s.remaining_demand, 6)
        self.assertEqual(s.arrival_offset, 5)
        self.assertEqual(s.remaining_time, 55)

    def test_kwarg_order(self) -> None:
        """ Tests that the order of overflow of kwargs for SessionInfo is maintained.
        """
        s = SessionInfo("PS-001", "01", 10, 4, 5, 60, 63, 6, [8.0] * 54, [32.0] * 54)
        self.assertEqual(s.estimated_departure, 63)
        self.assertEqual(s.current_time, 6)
        nptest.assert_array_equal(s.min_rates, 8.0)
        self.assertEqual(s.min_rates.shape, (54,))
        nptest.assert_array_equal(s.max_rates, 32.0)
        self.assertEqual(s.max_rates.shape, (54,))

    def test_valid_inputs_distinct_estimated_departure(self) -> None:
        s = SessionInfo("PS-001", "01", 10, 4, 5, 60, estimated_departure=63)
        self.assertEqual(s.departure, 60)
        self.assertEqual(s.estimated_departure, 63)

    def test_valid_inputs_nonzero_current_time_greater_than_arrival(self) -> None:
        s = SessionInfo(
            "PS-001", "01", 10, 4, 5, 60, estimated_departure=63, current_time=6
        )
        self.assertEqual(s.current_time, 6)
        self.assertEqual(s.arrival_offset, 0)
        self.assertEqual(s.remaining_time, 54)

    def test_valid_inputs_nonzero_current_time_less_than_arrival(self) -> None:
        s = SessionInfo(
            "PS-001", "01", 10, 4, 5, 60, estimated_departure=63, current_time=4
        )
        self.assertEqual(s.current_time, 4)
        self.assertEqual(s.arrival_offset, 1)
        self.assertEqual(s.remaining_time, 55)

    def test_proper_length_min_rates(self) -> None:
        s = SessionInfo(
            "PS-001", "01", 10, 4, 5, 10, estimated_departure=12, min_rates=[8.0] * 5
        )
        self.assertEqual(s.min_rates.shape, (5,))
        nptest.assert_array_equal(s.min_rates, 8.0)

    def test_min_rates_too_short(self) -> None:
        with self.assertRaises(ValueError):
            SessionInfo(
                "PS-001",
                "01",
                10,
                4,
                5,
                10,
                estimated_departure=12,
                min_rates=[8.0] * 4,
            )

    def test_min_rates_too_long(self) -> None:
        with self.assertRaises(ValueError):
            SessionInfo(
                "PS-001",
                "01",
                10,
                4,
                5,
                10,
                estimated_departure=12,
                min_rates=[8.0] * 6,
            )

    def test_proper_length_max_rates(self) -> None:
        s = SessionInfo(
            "PS-001", "01", 10, 4, 5, 10, estimated_departure=12, max_rates=[8.0] * 5
        )
        self.assertEqual(s.max_rates.shape, (5,))
        nptest.assert_array_equal(s.max_rates, 8.0)

    def test_max_rates_too_short(self) -> None:
        with self.assertRaises(ValueError):
            SessionInfo(
                "PS-001",
                "01",
                10,
                4,
                5,
                10,
                estimated_departure=12,
                max_rates=[8.0] * 4,
            )

    def test_max_rates_too_long(self) -> None:
        with self.assertRaises(ValueError):
            SessionInfo(
                "PS-001",
                "01",
                10,
                4,
                5,
                10,
                estimated_departure=12,
                max_rates=[8.0] * 6,
            )


class TestInfrastructureInfo(TestCase):
    def test_inputs_consistent(self) -> None:
        m, n = 6, 5
        # Here, the last two arguments are kwargs, left unspecified to test that the
        # order of overflow is unchanged.
        infra = InfrastructureInfo(
            np.ones((m, n)),
            np.ones((m,)),
            np.ones((n,)),
            np.ones((n,)),
            [f"C-{i}" for i in range(m)],
            [f"S-{i}" for i in range(n)],
            np.ones((n,)),
            np.zeros((n,)),
            [np.array([1, 2, 3, 4])] * n,  # allowable_pilots
            np.zeros((n,)),  # is_continuous
        )
        self.assertEqual(infra.constraint_matrix.shape, (m, n))
        self.assertEqual(infra.constraint_limits.shape, (m,))
        self.assertEqual(infra.phases.shape, (n,))
        self.assertEqual(infra.voltages.shape, (n,))
        self.assertEqual(len(infra.constraint_ids), m)
        self.assertEqual(len(infra.station_ids), n)
        self.assertEqual(infra.max_pilot.shape, (n,))
        self.assertEqual(infra.min_pilot.shape, (n,))
        self.assertEqual(len(infra.allowable_pilots), n)
        self.assertEqual(infra.is_continuous.shape, (n,))
        self.assertEqual(infra.is_continuous.dtype, "bool")

    def test_inputs_allowable_pilot_defaults(self) -> None:
        m, n = 6, 5
        infra = InfrastructureInfo(
            np.ones((m, n)),
            np.ones((m,)),
            np.ones((n,)),
            np.ones((n,)),
            [f"C-{i}" for i in range(m)],
            [f"S-{i}" for i in range(n)],
            np.ones((n,)),
            np.zeros((n,)),
            is_continuous=np.zeros((n,)),
        )
        self.assertEqual(len(infra.allowable_pilots), n)
        self.assertEqual(infra.allowable_pilots, [None] * n)

    def test_inputs_is_continuous_default(self) -> None:
        m, n = 6, 5
        infra = InfrastructureInfo(
            np.ones((m, n)),
            np.ones((m,)),
            np.ones((n,)),
            np.ones((n,)),
            [f"C-{i}" for i in range(m)],
            [f"S-{i}" for i in range(n)],
            np.ones((n,)),
            np.zeros((n,)),
            allowable_pilots=[np.array([1, 2, 3, 4])] * n,
        )
        nptest.assert_array_equal(infra.is_continuous, True)
        self.assertEqual(infra.is_continuous.shape, (n,))
        self.assertEqual(infra.is_continuous.dtype, "bool")

    def test_num_stations_mismatch(self) -> None:
        m, n = 5, 6
        for i in range(8):
            for error in [-1, 1]:
                errors = [0] * 8
                errors[i] = error
                with self.assertRaises(ValueError):
                    InfrastructureInfo(
                        np.ones((m, n + errors[0])),
                        np.ones((m,)),
                        np.ones((n + errors[1],)),
                        np.ones((n + errors[2],)),
                        [f"C-{i}" for i in range(m)],
                        [f"S-{i}" for i in range(n + errors[3])],
                        np.ones((n + errors[4],)),
                        np.zeros((n + errors[5],)),
                        allowable_pilots=[np.array([1, 2, 3, 4])] * (n + errors[6]),
                        is_continuous=np.zeros((n + errors[7],)),
                    )

    def test_num_constraints_mismatch(self) -> None:
        m, n = 5, 6
        for i in range(3):
            for error in [-1, 1]:
                errors = [0] * 3
                errors[i] = error
                with self.assertRaises(ValueError):
                    InfrastructureInfo(
                        np.ones((m + errors[0], n)),
                        np.ones((m + errors[1],)),
                        np.ones((n,)),
                        np.ones((n,)),
                        [f"C-{i}" for i in range(m + errors[2])],
                        [f"S-{i}" for i in range(n)],
                        np.ones((n,)),
                        np.zeros((n,)),
                        allowable_pilots=[np.array([1, 2, 3, 4])] * n,
                        is_continuous=np.zeros((n,)),
                    )

    def test_num_stations_num_constraints_mismatch(self) -> None:
        m, n = 5, 6
        with self.assertRaises(ValueError):
            InfrastructureInfo(
                np.ones((m + 1, n)),
                np.ones((m,)),
                np.ones((n,)),
                np.ones((n,)),
                [f"C-{i}" for i in range(m - 1)],
                [f"S-{i}" for i in range(n)],
                np.ones((n,)),
                np.zeros((n + 1,)),
                allowable_pilots=[np.array([1, 2, 3, 4])] * n,
                is_continuous=np.zeros((n - 1,)),
            )


class TestInterface(TestCase):
    def setUp(self) -> None:
        """ Run this setup function once before each test. """
        self.simulator = create_autospec(Simulator)
        self.network = ChargingNetwork()
        self.simulator.network = self.network
        self.interface = Interface(self.simulator)
        evse1 = EVSE("PS-001")
        self.network.register_evse(evse1, 120, -30)
        evse2 = EVSE("PS-002")
        evse3 = DeadbandEVSE("PS-003")
        self.network.register_evse(evse3, 360, 150)
        self.network.register_evse(evse2, 240, 90)
        # Include a FiniteRatesEVSE for more thorough testing.
        self.allowable_rates: List[int] = [0, 8, 16, 24, 32]
        evse4: FiniteRatesEVSE = FiniteRatesEVSE("PS-004", self.allowable_rates)
        self.network.register_evse(evse4, 120, -30)
        self.network.constraint_matrix = np.eye(4)
        self.network.magnitudes = np.ones((4, 1))
        self.network.constraint_index = ["C1", "C2", "C3", "C4"]

    def test_init(self) -> None:
        self.assertIs(self.interface._simulator, self.simulator)

    def test_violation_tolerance(self) -> None:
        self.assertEqual(
            self.interface._violation_tolerance,
            self.interface._simulator.network.violation_tolerance,
        )

    def test_relative_tolerance(self) -> None:
        self.assertEqual(
            self.interface._relative_tolerance,
            self.interface._simulator.network.relative_tolerance,
        )

    def test_active_evs(self) -> None:
        with self.assertWarns(UserWarning):
            _ = self.interface.active_evs
        self.simulator.get_active_evs.assert_called_once()

    def test_last_applied_pilot_signals_low_iteration(self) -> None:
        self.simulator.iteration = 1
        self.assertEqual(self.interface.last_applied_pilot_signals, {})

    def test_allowable_pilot_signals(self) -> None:
        self.assertEqual(
            self.interface.allowable_pilot_signals("PS-001"), (True, [0, float("inf")])
        )

    def test_allowable_pilot_signals_deadband(self) -> None:
        self.assertEqual(
            self.interface.allowable_pilot_signals("PS-003"), (True, [6, float("inf")])
        )

    def test_allowable_pilot_signals_finite_rates(self) -> None:
        self.assertEqual(
            self.interface.allowable_pilot_signals("PS-004"),
            (False, self.allowable_rates),
        )

    def test_max_pilot_signal(self) -> None:
        self.assertEqual(self.interface.max_pilot_signal("PS-001"), float("inf"))

    def test_max_pilot_signal_deadband(self) -> None:
        self.assertEqual(self.interface.max_pilot_signal("PS-003"), float("inf"))

    def test_max_pilot_signal_finite_rates(self) -> None:
        self.assertEqual(self.interface.max_pilot_signal("PS-004"), 32)

    def test_min_pilot_signal(self) -> None:
        self.assertEqual(self.interface.min_pilot_signal("PS-001"), 0)

    def test_min_pilot_signal_deadband(self) -> None:
        self.assertEqual(self.interface.min_pilot_signal("PS-003"), 0)

    def test_min_pilot_signal_finite_rates(self) -> None:
        self.assertEqual(self.interface.min_pilot_signal("PS-004"), 8)

    def test_evse_voltage(self) -> None:
        self.assertEqual(self.interface.evse_voltage("PS-002"), 240)

    def test_evse_phase(self) -> None:
        self.assertEqual(self.interface.evse_phase("PS-002"), 90)

    def test_get_constraints(self) -> None:
        constraint_info: Constraint = self.interface.get_constraints()
        nptest.assert_equal(
            constraint_info.constraint_matrix, self.network.constraint_matrix
        )
        nptest.assert_equal(constraint_info.magnitudes, self.network.magnitudes)
        self.assertEqual(
            constraint_info.constraint_index, self.network.constraint_index
        )
        self.assertEqual(constraint_info.evse_index, self.network.station_ids)

    def test_is_feasible_empty_schedule(self) -> None:
        self.assertTrue(self.interface.is_feasible({}))

    def test_is_feasible_unequal_schedules(self) -> None:
        with self.assertRaises(InvalidScheduleError):
            self.interface.is_feasible(
                {"PS-001": [1, 2], "PS-002": [3, 4, 5], "PS-003": [4, 5]}
            )

    def test_is_feasible(self) -> None:
        # Mock network's is_feasible function to check its call signature later
        self.network.is_feasible = create_autospec(self.network.is_feasible)
        self.interface.is_feasible({"PS-001": [1, 2], "PS-002": [4, 5]})
        network_is_feasible_args = self.network.is_feasible.call_args
        # Check that the call to the network's is_feasible method has the correct
        # arguments.
        np.testing.assert_allclose(
            network_is_feasible_args[0][0], np.array([[1, 2], [0, 0], [4, 5], [0, 0]])
        )
        # Network's is_feasible method has its second argument (linear) defaulting to
        # False. Check this is the case.
        self.assertEqual(network_is_feasible_args[0][1], False)
        # Network's is_feasible method has its third argument (violation_tolerance)
        # defaulting to None. Check this is the case.
        self.assertEqual(network_is_feasible_args[0][2], 1e-5)
        self.assertEqual(network_is_feasible_args[0][3], 1e-7)

    def test_is_feasible_with_options(self) -> None:
        # Mock network's is_feasible function to check its call signature later
        self.network.is_feasible = create_autospec(self.network.is_feasible)
        self.interface.is_feasible(
            {"PS-001": [1, 2], "PS-002": [4, 5]},
            linear=True,
            violation_tolerance=1e-3,
            relative_tolerance=1e-5,
        )
        network_is_feasible_args = self.network.is_feasible.call_args
        # Check that the call to the network's is_feasible method has the correct
        # arguments.
        np.testing.assert_allclose(
            network_is_feasible_args[0][0], np.array([[1, 2], [0, 0], [4, 5], [0, 0]])
        )
        self.assertEqual(network_is_feasible_args[0][1], True)
        self.assertEqual(network_is_feasible_args[0][2], 1e-3)
        self.assertEqual(network_is_feasible_args[0][3], 1e-5)



# coding=utf-8
"""
Abstract base algorithm for scheduled charging.
"""
import time
from typing import Optional, List, Dict, Union


class BaseAlgorithm:
    """ Abstract base class meant to be inherited from to implement new algorithms.

    Subclasses must implement the schedule method.

    Attributes:
        max_recompute (int): Maximum number of periods between calling the scheduling
            algorithm even if no events occur. If None, the scheduling algorithm is
            only called when an event occurs. Default: None.
    """

    _interface: Optional["Interface"]
    max_recompute: Optional[int]
    solve_stats: List[Dict[str, Union[float, int]]]

    def __init__(self) -> None:
        self._interface = None
        self.max_recompute = None
        self.solve_stats = []

    def __repr__(self) -> str:
        arg_str = ", ".join([f"{key}={value}" for key, value in self.__dict__.items()])
        return f"{self.__class__.__name__}({arg_str})"

    @property
    def interface(self) -> "Interface":
        """ Return the algorithm's interface with the environment.

        Returns:
            Interface: An interface to the environment.

        Raises:
            ValueError: Exception raised if interface is accessed prior to an interface
                being registered.
        """
        if self._interface is not None:
            return self._interface
        else:
            raise ValueError(
                "No interface has been registered yet. Please call register_interface "
                "prior to using the algorithm."
            )

    def register_interface(self, interface: "Interface") -> None:
        """ Register interface to the _simulator/physical system.

        This interface is the only connection between the algorithm and what it is
        controlling. Its purpose is to abstract the underlying network so that the same
        algorithms can run on a simulated environment or a physical one.

        Args:
            interface (Interface): An interface to the underlying network whether
            simulated or real.

        Returns:
            None
        """
        self._interface = interface

    def schedule(self, active_sessions: List["SessionInfo"]) -> Dict[str, List[float]]:
        """ Creates a schedule of charging rates for each ev in the active_evs list.

        NOT IMPLEMENTED IN BaseAlgorithm. This method MUST be implemented in all
        subclasses.

        This method returns a schedule of charging rates for each
        Args:
            active_sessions (List[SessionInfo]): List of SessionInfo objects
                which are currently ready to be charged and not finished
                charging.

        Returns:
            Dict[str, List[float]]: Dictionary mapping a station_id to a list of
                charging rates. Each charging rate is valid for one period measured
                relative to the current period, i.e. schedule['abc'][0] is the
                charging rate for station 'abc' during the current period and
                schedule['abc'][1] is the charging rate for the next period,
                and so on. If an algorithm only produces charging rates for the
                current time period, the length of each list should be 1. If this is
                the case, make sure to also set the maximum resolve period to be 1
                period so that the algorithm will be called each period. An
                alternative is to repeat the charging rate a number of times equal to
                the max recompute period.
        """
        raise NotImplementedError

    def run(self) -> Dict[str, List[float]]:
        """ Runs the scheduling algorithm for the current period and returns the
        resulting schedules.

        Returns:
            See schedule.
        """
        active_sessions = self.interface.active_sessions()
        start_time = time.perf_counter()
        schedules = self.schedule(active_sessions)
        solve_time = time.perf_counter() - start_time
        self.solve_stats.append(
            {"solve_time": solve_time, "active_sessions": len(active_sessions)}
        )
        return schedules



# coding=utf-8
"""
Tests provided sorting algorithms under many cases.
"""
import random
import unittest
from unittest.mock import Mock
from numpy import testing as nptest
from collections import namedtuple

from acnportal.algorithms import *
from acnportal.algorithms.tests.generate_test_cases import *
from acnportal.algorithms.tests.testing_interface import TestingInterface
from acnportal.algorithms import UpperBoundEstimatorBase

CURRENT_TIME = 0
PERIOD = 5

# -----------------------------------------------------------------------------
# Algorithms to Test
# -----------------------------------------------------------------------------
algorithms = {
    "FCFS": SortedSchedulingAlgo(first_come_first_served),
    "LLF": SortedSchedulingAlgo(least_laxity_first),
    "EDF": SortedSchedulingAlgo(earliest_deadline_first),
    "LCFS": SortedSchedulingAlgo(last_come_first_served),
    "LRPT": SortedSchedulingAlgo(largest_remaining_processing_time),
    "RR": RoundRobin(first_come_first_served),
}

# -----------------------------------------------------------------------------
# Test Suite
# -----------------------------------------------------------------------------
Scenario = namedtuple(
    "Scenario",
    ["name", "interface", "assert_at_max", "uninterrupted", "estimate_max_rate"],
)


class BaseAlgorithmTest(unittest.TestCase):
    def setUp(self) -> None:
        """
        Tests that a given algorithm provides feasible schedules to a simulation.
        The elements of feasibility tested here are:

        - A given charging rate is <= the maximum rate of the EVSE it is sent to.

        - A given charging rate is <= the maximum rate of the Session it is charging.

        - A given charging rate is in the allowable rate set of the Session it is
        charging.

        - No session is given more energy than it requested.

        - Infrastructure limits are satisfied.

        - TODO: what is the function of the assert at max rate?

        - Charging is uninterrupted (never goes to zero during the session unless the
        vehicle is done charging) if required.

        - A max rate estimation, if provided, is not exceeded during the session.

        Each algorithm test class has an algorithm (set by overriding this function
        and setting an actual algorithm) and a max_rate_estimation, which provides the
        return value of a mocked UpperBoundEstimator (used if max rate estimation is
        tested).

        The implementation of the _get_scenarios method details which charging
        scenarios should be run in this test class. A scenario is defined by a name,
        interface (usually a testing interface with static simulator data,
        see TestingInterface), and attributes assert_at_max, uninterrupted,
        and estimate_max rate, which dictate additional constraints under which the
        algorithm should operate.

        Returns:
            None.
        """
        self.algo = None
        self.max_rate_estimation = {}

    @staticmethod
    def _get_scenarios() -> List[Scenario]:
        return []

    def test_output_feasible(self) -> None:
        scenarios = self._get_scenarios()
        for scenario in scenarios:
            self.algo.register_interface(scenario.interface)
            self.algo.uninterrupted = scenario.uninterrupted
            estimator_mock = UpperBoundEstimatorBase()
            estimator_mock.get_maximum_rates = Mock(
                return_value=self.max_rate_estimation
            )
            self.algo.max_rate_estimator = estimator_mock
            self.algo.estimate_max_rate = scenario.estimate_max_rate

            schedule = self.algo.run()
            self._run_tests(
                scenario.name,
                scenario.interface.active_sessions(),
                schedule,
                scenario.interface,
                scenario.assert_at_max,
                scenario.uninterrupted,
            )

    def _run_tests(
        self,
        name: str,
        sessions: List[SessionInfo],
        schedule: Dict[str, List[float]],
        interface: Interface,
        assert_at_max: bool = False,
        uninterrupted: bool = False,
    ) -> None:
        with self.subTest(msg=f"test_all_rates_less_than_evse_limit - {name}"):
            self._test_all_rates_less_than_evse_limit(schedule, interface)

        with self.subTest(msg=f"test_all_rates_less_than_session_max_rates - {name}"):
            self._test_all_rates_less_than_session_max_rates(sessions, schedule)

        with self.subTest(
            msg=f"test_all_rates_greater_than_session_min_rates - {name}"
        ):
            self._test_all_rates_greater_than_session_min_rates(
                sessions, interface, schedule
            )

        with self.subTest(f"test_in_allowable_rates - {name}"):
            self._test_in_allowable_rates(sessions, schedule, interface)

        with self.subTest(f"test_energy_requested_not_exceeded - {name}"):
            self._test_energy_requested_not_exceeded(sessions, schedule, interface)

        with self.subTest(f"test_infrastructure_limits_satisfied - {name}"):
            self._test_infrastructure_limits_satisfied(schedule, interface)

        if assert_at_max:
            with self.subTest(f"test_all_rates_at_max - {name}"):
                self._test_all_rates_at_max(sessions, schedule, interface)

        if uninterrupted:
            with self.subTest(f"test_charging_not_interrupted - {name}"):
                self._test_charging_not_interrupted(sessions, schedule, interface)

        if self.algo.estimate_max_rate:
            with self.subTest(f"test_max_rate_estimator_not_exceeded - {name}"):
                self._test_max_rate_estimator_not_exceeded(sessions, schedule)

    def _test_all_rates_less_than_evse_limit(self, schedule, interface) -> None:
        for station_id, rates in schedule.items():
            self.assertLessEqual(rates, interface.max_pilot_signal(station_id))

    def _test_all_rates_less_than_session_max_rates(self, sessions, schedule) -> None:
        for session in sessions:
            station_id = session.station_id
            self.assertLessEqual(schedule[station_id][0], session.max_rates[0])

    def _test_all_rates_greater_than_session_min_rates(
        self, sessions, interface, schedule
    ) -> None:
        infrastructure = interface.infrastructure_info()
        for session in sessions:
            station_id = session.station_id
            station_index = infrastructure.get_station_index(session.station_id)
            threshold = (
                infrastructure.min_pilot[station_index]
                * infrastructure.voltages[station_index]
                / (60 / interface.period)
                / 1000
            )
            if session.remaining_demand > threshold:
                self.assertGreaterEqual(schedule[station_id][0], session.min_rates[0])
            else:
                self.assertEqual(schedule[station_id][0], 0)

    def _test_in_allowable_rates(self, sessions, schedule, interface) -> None:
        for session in sessions:
            station_id = session.station_id
            (is_continuous, allowable,) = interface.allowable_pilot_signals(station_id)
            if is_continuous:
                self.assertGreaterEqual(schedule[station_id], allowable[0])
                self.assertLessEqual(schedule[station_id], allowable[1])
            else:
                self.assertIn(schedule[station_id], allowable)

    def _test_energy_requested_not_exceeded(
        self, sessions, schedule, interface
    ) -> None:
        for session in sessions:
            station_id = session.station_id
            self.assertLessEqual(
                schedule[station_id], interface.remaining_amp_periods(session),
            )

    def _test_infrastructure_limits_satisfied(self, schedule, interface) -> None:
        self.assertTrue(interface.is_feasible(schedule))

    # noinspection PyMethodMayBeStatic
    def _test_all_rates_at_max(
        self, sessions, schedule, interface
    ) -> None:  # pylint: disable=no-self-use
        infrastructure = interface.infrastructure_info()
        for session in sessions:
            i = infrastructure.get_station_index(session.station_id)
            ub = min(infrastructure.max_pilot[i], session.max_rates[0])
            rates = schedule[session.station_id]
            nptest.assert_almost_equal(rates, ub, decimal=4)

    def _test_charging_not_interrupted(self, sessions, schedule, interface) -> None:
        for session in sessions:
            scheduled = schedule[session.station_id]
            minimum_pilot = interface.min_pilot_signal(session.station_id)
            remaining_energy = interface.remaining_amp_periods(session)

            # Algorithm should not exceed remaining energy in order to meet minimum
            # pilot.
            if minimum_pilot < remaining_energy:
                self.assertGreaterEqual(scheduled, minimum_pilot)

    def _test_max_rate_estimator_not_exceeded(self, sessions, schedule) -> None:
        for session in sessions:
            self.assertLessEqual(
                np.array(schedule[session.station_id]),
                self.max_rate_estimation[session.session_id],
            )


# Two Station Test Case
def two_station(
    limit: float,
    continuous: bool,
    session_max_rate: float,
    session_min_rate: float = 0,
    remaining_energy: Optional[List[float]] = None,
    estimated_departure: Optional[List[float]] = None,
) -> TestingInterface:
    """ Two EVSEs with the same phase, one constraint, and allowable rates from 0 to 32
    if continuous; integers between 8 and 32 if not. Also provides 2 sessions arriving
    and departing at the same time, with the same energy demands. """
    if continuous:
        allowable: List[np.ndarray] = [np.array([0, 32])] * 2
    else:
        allowable: List[np.ndarray] = [np.array([0] + list(range(8, 33)))] * 2
    if remaining_energy is None:
        remaining_energy: List[float] = [3.3, 3.3]
    network: InfrastructureDict = single_phase_single_constraint(
        2, limit, allowable_pilots=allowable, is_continuous=np.array([continuous] * 2)
    )
    sessions: List[SessionDict] = session_generator(
        num_sessions=2,
        arrivals=[0] * 2,
        departures=[11, 12],
        estimated_departures=estimated_departure,
        requested_energy=[3.3] * 2,
        remaining_energy=remaining_energy,
        min_rates=[session_min_rate] * 2,
        max_rates=[session_max_rate] * 2,
    )
    data = {
        "active_sessions": sessions,
        "infrastructure_info": network,
        "current_time": CURRENT_TIME,
        "period": PERIOD,
    }
    return TestingInterface(data)


def big_three_phase_network(
    num_sessions: int = 30, limit: float = 1000
) -> TestingInterface:
    """
    Network is WAY over designed to deal with unbalance and ensure all EVs can charge
    at their maximum rate.
    """
    network = three_phase_balanced_network(num_sessions // 3, limit)
    sessions = session_generator(
        num_sessions=num_sessions,
        arrivals=[0] * num_sessions,
        departures=[2] * num_sessions,
        requested_energy=[10] * num_sessions,
        remaining_energy=[10] * num_sessions,
        max_rates=[32] * num_sessions,
    )
    random.shuffle(sessions)
    data = {
        "active_sessions": sessions,
        "infrastructure_info": network,
        "current_time": CURRENT_TIME,
        "period": PERIOD,
    }
    return TestingInterface(data)


class TestTwoStationsBase(BaseAlgorithmTest):
    def setUp(self) -> None:
        """ See BaseAlgorithmTest.setUp. """
        self.algo = None
        self.max_rate_estimation = {"0": 16, "1": 12}

    @staticmethod
    def _get_scenarios() -> List[Scenario]:
        scenarios = []
        # Test one limit where constraints are binding (40 A),
        # one where constraint will be met exactly by charging at max (64 A),
        # and one where constraints are not binding at all (80 A).
        for limit in [40, 64, 80]:
            # To ensure that both station limits and session limits are
            # satisfied, consider several max_rate parameters.
            # At 16 A, session limit is below station limit (32)
            # At 32 A, session limit equals station limit
            # At 40 A, session limit is greater than station limit.
            for session_max_rate in [16, 32, 40]:
                # Consider both continuous and discrete pilot signals
                for session_min_rate in [0, 8]:
                    # The latter case below tests when the remaining amp periods is
                    # small enough to trigger a pilot signal going to 0 while there is
                    # still demand remaining.
                    for session_energy_demands in [[3.3, 3.3], [0.3, 0.05]]:
                        # Consider continuous and discrete EVSEs
                        for continuous in [True, False]:
                            # Consider both interruptable and uninterrupted charging
                            for uninterrupted in [True, False]:
                                for estimate_max_rate in [True, False]:
                                    if (
                                        limit < 64
                                        or session_energy_demands == [0.3, 0.05]
                                        or estimate_max_rate
                                    ):
                                        assert_at_max = False
                                    else:
                                        assert_at_max = True
                                    interface: TestingInterface = two_station(
                                        limit,
                                        continuous,
                                        session_max_rate,
                                        session_min_rate,
                                        remaining_energy=session_energy_demands,
                                    )
                                    scenario_name = (
                                        f"capacity: {limit}, "
                                        f"session max: {session_max_rate}, "
                                        f"session min: {session_min_rate}, "
                                        f"continuous pilot: {continuous}, "
                                        f"uninterrupted: {uninterrupted}, "
                                        f"estimate_max_rate: {estimate_max_rate} "
                                    )
                                    scenarios.append(
                                        Scenario(
                                            scenario_name,
                                            interface,
                                            assert_at_max,
                                            uninterrupted,
                                            estimate_max_rate,
                                        )
                                    )
        return scenarios


class TestThirtyStationsBase(BaseAlgorithmTest):
    def setUp(self) -> None:
        """ See BaseAlgorithmTest.setUp. """
        self.algo = None
        self.max_rate_estimation = {}  # Don't use max_rate_estimation for this test.

    @staticmethod
    def _get_scenarios() -> List[Scenario]:
        scenarios = []
        for limit in [1500, 3200]:
            interface: TestingInterface = big_three_phase_network(limit=limit)
            scenario_name = f"capacity: {limit} "
            scenarios.append(Scenario(scenario_name, interface, False, False, False))
        return scenarios


# --------------------------------------------------------------------------------------
# Tests
#
# As the functionality tested in each class is evident from the class names,
# the setUp methods are left without docstrings (hence the noinspection
# comments).
# --------------------------------------------------------------------------------------
class TestTwoStationsMinRatesInfeasible(unittest.TestCase):
    """ Check that error is thrown when minimum rates are not feasible. """

    def test_sorted_min_rates_infeasible(self) -> None:
        limit = 16
        max_rate = 32
        min_rate = 16
        for continuous in [True, False]:
            interface: TestingInterface = two_station(
                limit, continuous, max_rate, min_rate
            )
            for algo_name, algo in algorithms.items():
                scenario_name = (
                    f"algorithm: {algo_name}, "
                    f"capacity: {limit}, "
                    f"continuous pilot: {continuous}"
                )
                with self.subTest(msg=f"{scenario_name}"):
                    algo.register_interface(interface)
                    with self.assertRaisesRegex(
                        ValueError,
                        "Charging all sessions at "
                        "their lower bound is not "
                        "feasible.",
                    ):
                        algo.run()


class TestTwoStationsFCFS(TestTwoStationsBase):
    # noinspection PyMissingOrEmptyDocstring
    def setUp(self) -> None:
        super().setUp()
        self.algo = SortedSchedulingAlgo(first_come_first_served)


class TestThirtyStationsFCFS(TestThirtyStationsBase):
    # noinspection PyMissingOrEmptyDocstring
    def setUp(self) -> None:
        super().setUp()
        self.algo = SortedSchedulingAlgo(first_come_first_served)


class TestTwoStationsEDF(TestTwoStationsBase):
    # noinspection PyMissingOrEmptyDocstring
    def setUp(self) -> None:
        super().setUp()
        self.algo = SortedSchedulingAlgo(earliest_deadline_first)


class TestThirtyStationsEDF(TestThirtyStationsBase):
    # noinspection PyMissingOrEmptyDocstring
    def setUp(self) -> None:
        super().setUp()
        self.algo = SortedSchedulingAlgo(earliest_deadline_first)


class TestTwoStationsLLF(TestTwoStationsBase):
    # noinspection PyMissingOrEmptyDocstring
    def setUp(self) -> None:
        super().setUp()
        self.algo = SortedSchedulingAlgo(least_laxity_first)


class TestThirtyStationsLLF(TestThirtyStationsBase):
    # noinspection PyMissingOrEmptyDocstring
    def setUp(self) -> None:
        super().setUp()
        self.algo = SortedSchedulingAlgo(least_laxity_first)


class TestTwoStationsLCFS(TestTwoStationsBase):
    # noinspection PyMissingOrEmptyDocstring
    def setUp(self) -> None:
        super().setUp()
        self.algo = SortedSchedulingAlgo(last_come_first_served)


class TestThirtyStationsLCFS(TestThirtyStationsBase):
    # noinspection PyMissingOrEmptyDocstring
    def setUp(self) -> None:
        super().setUp()
        self.algo = SortedSchedulingAlgo(last_come_first_served)


class TestTwoStationsLRPT(TestTwoStationsBase):
    # noinspection PyMissingOrEmptyDocstring
    def setUp(self) -> None:
        super().setUp()
        self.algo = SortedSchedulingAlgo(largest_remaining_processing_time)


class TestThirtyStationsLRPT(TestThirtyStationsBase):
    # noinspection PyMissingOrEmptyDocstring
    def setUp(self) -> None:
        super().setUp()
        self.algo = SortedSchedulingAlgo(largest_remaining_processing_time)


class TestTwoStationsRR(TestTwoStationsBase):
    # noinspection PyMissingOrEmptyDocstring
    def setUp(self) -> None:
        super().setUp()
        self.algo = RoundRobin(first_come_first_served)


class TestThirtyStationsRR(TestThirtyStationsBase):
    # noinspection PyMissingOrEmptyDocstring
    def setUp(self) -> None:
        super().setUp()
        self.algo = RoundRobin(first_come_first_served)

class TestEarliestDeadlineFirstOrder(unittest.TestCase):
    def test_estimated_departure_matches_real(self):
        interface = two_station(limit=100, continuous=True, session_max_rate=32, estimated_departure=[11, 12])
        sorted_sessions = earliest_deadline_first(interface.active_sessions(), interface)
        self.assertEqual(sorted_sessions[0].session_id, "0")
        self.assertEqual(sorted_sessions[1].session_id, "1")


    def test_estimated_departure_does_not_match_real(self):
        interface = two_station(limit=100, continuous=True, session_max_rate=32, estimated_departure=[12, 11])
        sorted_sessions = earliest_deadline_first(interface.active_sessions(), interface)
        self.assertEqual(sorted_sessions[0].session_id, "1")
        self.assertEqual(sorted_sessions[1].session_id, "0")


class TestLeastLaxityFirstOrder(unittest.TestCase):
    def test_estimated_departure_matches_real(self):
        interface = two_station(limit=100, continuous=True, session_max_rate=32, estimated_departure=[11, 12])
        sorted_sessions = least_laxity_first(interface.active_sessions(), interface)
        self.assertEqual(sorted_sessions[0].session_id, "0")
        self.assertEqual(sorted_sessions[1].session_id, "1")


    def test_estimated_departure_does_not_match_real(self):
        interface = two_station(limit=100, continuous=True, session_max_rate=32, estimated_departure=[12, 11])
        sorted_sessions = least_laxity_first(interface.active_sessions(), interface)
        self.assertEqual(sorted_sessions[0].session_id, "1")
        self.assertEqual(sorted_sessions[1].session_id, "0")


del BaseAlgorithmTest
del TestTwoStationsBase
del TestThirtyStationsBase
# del TestTwoStationsMinRatesFeasibleBase



import numpy as np
import warnings
import datetime


def aggregate_current(sim):
    """ Calculate the time series of aggregate current of all EVSEs within a simulation.

    Args:
        sim (Simulator): A Simulator object which has been run.

    Returns:
        np.Array: A numpy ndarray of the aggregate current at each time. [A]
    """
    return sim.charging_rates.sum(axis=0)


def aggregate_power(sim):
    """ Calculate the time series of aggregate power of all EVSEs within a simulation.

    Args:
        sim (Simulator): A Simulator object which has been run.

    Returns:
        np.Array: A numpy ndarray of the aggregate power at each time. [kW]
    """
    return sim.network._voltages.T.dot(sim.charging_rates) / 1000


def constraint_currents(sim, return_magnitudes=False, constraint_ids=None):
    """ Calculate the time series of current for each constraint in the ChargingNetwork for a simulation.

    Args:
        sim (Simulator): A Simulator object which has been run.
        return_magnitudes (bool): If true, return constraint currents as real magnitudes instead of complex numbers.
        constraint_ids (List[str]): List of constraint names for which the current should be returned. If None, return
            all constraint currents.

    Returns:
        Dict (str, np.Array): A dictionary mapping the name of a constraint to a numpy array of the current subject to
            that constraint at each time.
    """
    if constraint_ids is None:
        constraint_ids = sim.network.constraint_index

    currents_list = sim.network.constraint_current(
        sim.charging_rates, constraints=constraint_ids
    )

    if not return_magnitudes:
        currents_list = np.abs(currents_list)
    # Ensure constraint_ids have correct order relative to constraint_index in network
    constraint_ids = [
        constraint_id
        for constraint_id in sim.network.constraint_index
        if constraint_id in constraint_ids
    ]

    return {constraint_ids[i]: currents_list[i] for i in range(len(constraint_ids))}


def proportion_of_energy_delivered(sim):
    """ Calculate the percentage of total energy delivered over total energy requested.

    Args:
        sim (Simulator): A Simulator object which has been run.

    Returns:
        float: Proportion of total energy requested which was delivered during the simulation.
    """
    total_requested = sum(ev.requested_energy for ev in sim.ev_history.values())
    total_delivered = total_energy_delivered(sim)
    return total_delivered / total_requested


def total_energy_delivered(sim):
    """ Calculate total energy delivered in kWh.

    Args:
        sim (Simulator): A Simulator object which has been run.

    Returns:
        float: Total energy delivered during the simulation [kWh]
    """
    total_delivered = sum(ev.energy_delivered for ev in sim.ev_history.values())
    return total_delivered


def total_energy_requested(sim):
    """ Calculate total energy requested in kWh.

    Args:
        sim (Simulator): A Simulator object which has been run.

    Returns:
        float: Total energy requested during the simulation [kWh]
    """
    total_requested = sum(ev.requested_energy for ev in sim.ev_history.values())
    return total_requested


def proportion_of_demands_met(sim, threshold=0.1):
    """ Calculate the percentage of charging sessions where the energy request was met.

    Args:
        sim (Simulator): A Simulator object which has been run.
        threshold (float): Close to finished a session should be to be considered finished. Default: 0.1. [kW]

    Returns:
        float: Proportion of sessions where the energy demand was fully met.
    """
    finished = sum(
        1 for ev in sim.ev_history.values() if ev.remaining_demand < threshold
    )
    return finished / len(sim.ev_history)


def current_unbalance(sim, phase_ids, unbalance_type="NEMA", type=None):
    """ Calculate the current unbalance for each time in simulation.

    Supports two definitions of unbalance.
    1)  The NEMA definition defined as the ratio of the maximum deviation of an RMS current from the average RMS current
        over the average RMS current.
            (max(|I_a|, |I_b|, |I_c|) - 1/3 (|I_a| + |I_b| + |I_c|)) / (1/3 (|I_a| + |I_b| + |I_c|))

    See https://www.powerstandards.com/Download/Brief%20Discussion%20of%20Unbalance%20Definitions.pdf for more info.

    Args:
        sim (Simulator): A Simulator object which has been run.
        phase_ids (List[str]): List of length 3 where each element is the identifier of phase A, B, and C respectively.
        unbalance_type (str): Method to use for calculating phase unbalance. Acceptable values are 'NEMA'.

    Returns:
        List[float]: Time series of current unbalance as a list with one value per timestep.
    """
    if type is not None:
        warnings.warn(
            "current_unbalance kwarg 'type' is deprecated. "
            "Use 'unbalance_type' instead.",
            DeprecationWarning,
        )
        unbalance_type = type
    if unbalance_type == "NEMA":
        return _nema_current_unbalance(sim, phase_ids)
    else:
        raise ValueError("type must be NEMA not {0}".format(unbalance_type))


def _nema_current_unbalance(sim, phase_ids):
    """ Calculate the current unbalance using the NEMA definition.

    The NEMA definition defined as the ratio of the maximum deviation of an RMS current from the average RMS current
    over the average RMS current.
        (max(|I_a|, |I_b|, |I_c|) - 1/3 (|I_a| + |I_b| + |I_c|)) / (1/3 (|I_a| + |I_b| + |I_c|))

    Args:
        sim (Simulator): A Simulator object which has been run.
        phase_ids (List[str]): List of length 3 where each element is the identifier of phase A, B, and C respectively.

    Returns:
        List[float]: Time series of current unbalance as a list with one value per timestep.
    """
    currents_dict = constraint_currents(sim, constraint_ids=phase_ids)
    currents = np.vstack([currents_dict[phase] for phase in phase_ids])
    return (np.max(currents, axis=0) - np.mean(currents, axis=0)) / np.mean(
        currents, axis=0
    )


def energy_cost(sim, tariff=None):
    """ Calculate the total energy cost of the simulation.

    Args:
        sim (Simulator): A Simulator object which has been run.
        tariff (TimeOfUseTariff): Tariff structure to use when calculating energy costs.

    Returns:
        float: Total energy cost of the simulation ($)

    """
    if tariff is None:
        if "tariff" in sim.signals:
            tariff = sim.signals["tariff"]
        else:
            raise ValueError("No pricing method is specified.")
    agg = aggregate_power(sim)
    energy_costs = tariff.get_tariffs(sim.start, len(agg), sim.period)
    return np.array(energy_costs).dot(agg) * (sim.period / 60)


def demand_charge(sim, tariff=None):
    """ Calculate the total demand charge of the simulation.

    Note this is only an accurate depiction of true demand charge if the simulation is exactly one billing period
    long.

    Args:
        sim (Simulator): A Simulator object which has been run.
        tariff (TimeOfUseTariff): Tariff structure to use when calculating energy costs.

    Returns:
        float: Total demand charge incurred by the simulation ($)

    """
    if tariff is None:
        if "tariff" in sim.signals:
            tariff = sim.signals["tariff"]
        else:
            raise ValueError("No pricing method is specified.")
    agg = aggregate_power(sim)
    dc = tariff.get_demand_charge(sim.start)
    return dc * np.max(agg)


def datetimes_array(sim):
    """ Return a numpy array of datetimes over which the simulation was
    run.

    The resolution of the datetimes list is equal to the period of the
    simulation, and the number of datetimes in the returned list is
    equal to the number of iterations of the simulation.

    Note that timezone information is not included with the datetime array.

    Args:
        sim (Simulator): A Simulator object which has been run.

    Returns:
        np.ndarray[np.datetime64]: List of datetimes over which the
            simulation was run.

    Warns:
        UserWarning: If this is called before sim is complete.
    """
    if not sim.event_queue.empty():
        warnings.warn(
            "Simulation incomplete; not all datetimes will be included.", UserWarning
        )
    no_tz_start: datetime = sim.start.replace(tzinfo=None)
    return np.array(
        [
            np.datetime64(no_tz_start + datetime.timedelta(minutes=sim.period * i))
            for i in range(sim.iteration)
        ]
    )



# coding=utf-8
"""
Defines several classes of Events in the simulation.
"""
from typing import Optional, Dict, Any, Tuple

from ..base import BaseSimObj
import warnings

from ..models.ev import EV


class Event(BaseSimObj):
    """ Base class for all events.

    Args:
        timestamp (int): Timestamp when an event occurs (periods)

    Attributes:
        timestamp (int): See args.
        event_type (str): Name of the event type.
        precedence (float): Used to order occurrence for events that happen in the same
            timestep. Higher precedence events occur before lower precedence events.

    """

    timestamp: int
    event_type: str
    precedence: float

    def __init__(self, timestamp: int) -> None:
        self.timestamp = timestamp
        self.event_type = ""
        self.precedence = float("inf")

    def __lt__(self, other: "Event") -> bool:
        """ Return True if the precedence of self is less than that of other.

        Args:
            other (Event like): Another Event-like object.

        Returns:
            bool
        """
        return self.precedence < other.precedence

    @property
    def type(self) -> str:
        """
        Legacy accessor for event_type. This will be removed in a future
        release.
        """
        warnings.warn(
            "Accessor 'type' for type of Event is deprecated. "
            "Use 'event_type' instead.",
            DeprecationWarning,
        )
        return self.event_type

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """ Implements BaseSimObj._to_dict. """
        attribute_dict = {
            "timestamp": self.timestamp,
            "event_type": self.event_type,
            "precedence": self.precedence,
        }
        return attribute_dict, context_dict

    @classmethod
    def _from_dict_helper(
        cls, out_obj: "Event", attribute_dict: Dict[str, Any]
    ) -> None:
        out_obj.event_type = attribute_dict["event_type"]
        out_obj.precedence = attribute_dict["precedence"]

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """ Implements BaseSimObj._from_dict. """
        out_obj = cls(attribute_dict["timestamp"])
        cls._from_dict_helper(out_obj, attribute_dict)
        return out_obj, loaded_dict


class EVEvent(Event):
    """ Subclass of Event for events which deal with an EV such as Plugin and Unplug
    events.

    Args:
        timestamp (int): See Event.
        ev (EV): The EV associated with this event.
    """

    ev: EV

    def __init__(self, timestamp: int, ev: EV) -> None:
        super().__init__(timestamp)
        self.ev = ev

    @property
    def station_id(self) -> str:
        """ Return the station_id for the EV associated with this Event. """
        return self.ev.station_id

    @property
    def session_id(self) -> str:
        """ Return the session_id for the EV associated with this Event. """
        return self.ev.session_id

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """ Implements BaseSimObj._to_dict. """
        attribute_dict, context_dict = super()._to_dict(context_dict)
        # Plugin-specific attributes

        # noinspection PyProtectedMember
        registry, context_dict = self.ev._to_registry(context_dict=context_dict)
        attribute_dict["ev"] = registry["id"]

        return attribute_dict, context_dict

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """ Implements BaseSimObj._from_dict. """
        # noinspection PyProtectedMember
        ev, loaded_dict = BaseSimObj._build_from_id(
            attribute_dict["ev"], context_dict, loaded_dict=loaded_dict
        )
        out_obj = cls(attribute_dict["timestamp"], ev)
        cls._from_dict_helper(out_obj, attribute_dict)
        return out_obj, loaded_dict


class PluginEvent(EVEvent):
    """ Subclass of Event for EV plugins.

    Args:
        timestamp (int): See Event.
        ev (EV): The EV which will be plugged in.
    """

    def __init__(self, timestamp: int, ev: EV) -> None:
        super().__init__(timestamp, ev)
        self.event_type = "Plugin"
        self.precedence = 10


class UnplugEvent(EVEvent):
    """ Subclass of Event for EV unplugs.

    Args:
        timestamp (int): See Event.
        ev (EV): The EV which will be unplugged.
    """

    def __init__(self, timestamp: int, ev: EV) -> None:
        super().__init__(timestamp, ev)
        self.event_type = "Unplug"
        self.precedence = 0

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        """ Implements BaseSimObj._from_dict. """
        # noinspection PyProtectedMember
        try:
            ev, loaded_dict = BaseSimObj._build_from_id(
                attribute_dict["ev"], context_dict, loaded_dict=loaded_dict
            )
        except KeyError:
            # In acnportal v0.2.2, UnplugEvent had session_id and station_id attributes
            # instead of an ev attribute. For backwards compatibility with previously
            # serialized UnplugEvents, we build the EV partially (including session and
            # station ids only) and set it as the ev attribute of the current
            # implementation of UnplugEvent.
            warnings.warn(
                "Loading UnplugEvents from an older version of acnportal into a newer "
                "one. UnplugEvent EV object will be incompletely deserialized."
            )
            ev = EV(
                -1,
                -1,
                -1,
                attribute_dict["station_id"],
                attribute_dict["session_id"],
                None,
            )
            for attribute in [
                "arrival",
                "departure",
                "requested_energy",
                "estimated_departure",
                "battery",
                "energy_delivered",
                "current_charging_rate",
            ]:
                delattr(ev, f"_{attribute}")
        out_obj = cls(attribute_dict["timestamp"], ev)
        cls._from_dict_helper(out_obj, attribute_dict)
        return out_obj, loaded_dict


class RecomputeEvent(Event):
    """ Subclass of Event for when the algorithm should be recomputed."""

    def __init__(self, timestamp: int) -> None:
        super().__init__(timestamp)
        self.event_type = "Recompute"
        self.precedence = 20



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
setuptools.setup(
    name="acnportal",
    version="0.3.2",
    author="Zachary Lee, Sunash Sharma",
    author_email="ev-help@caltech.edu",
    url="https://github.com/zach401/acnportal",
    description="A package of tools for large-scale EV charging research.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    package_data={
        "": ["LICENSE.txt", "THANKS.txt",],
        "acnportal": ["signals/tariffs/tariff_schedules/*"],
    },
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy",
        "pandas >= 1.1.0, < 1.2.0",
        "matplotlib",
        "requests",
        "pytz",
        "typing_extensions",
        "scikit-learn"
    ],
    extras_require={"all": ["scikit-learn"], "scikit-learn": ["scikit-learn"]},
)



""" This file contains extensions of ACN-Sim for testing purposes. """
from typing import Optional, Dict, Any, Tuple

from acnportal import acnsim
from acnportal.acnsim.base import BaseSimObj


class NamedEvent(acnsim.Event):
    """ An extension of Event that has a name. """

    def __init__(self, timestamp, name):
        super().__init__(timestamp)
        self.name = name


class DefaultNamedEvent(acnsim.Event):
    """ An extension of Event that has a name with a default. """

    def __init__(self, timestamp, name="my_event"):
        super().__init__(timestamp)
        self.name = name


class SetAttrEvent(acnsim.Event):
    """ An extension of Event with a settable attribute. """

    def __init__(self, timestamp):
        super().__init__(timestamp)

    def set_extra_attr(self, attr_val):
        # noinspection PyAttributeOutsideInit
        self.extra_attr = attr_val


class BatteryListEvent(acnsim.Event):
    """ An extension of Event with a list of Batteries. """

    def __init__(self, timestamp, battery_list):
        super().__init__(timestamp)
        self.battery_list = battery_list

    def _to_dict(
        self, context_dict: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        attribute_dict, context_dict = super()._to_dict(context_dict)
        battery_list = []
        for ev in self.battery_list:
            # noinspection PyProtectedMember
            registry, context_dict = ev._to_registry(context_dict=context_dict)
            battery_list.append(registry["id"])
        attribute_dict["battery_list"] = battery_list
        return attribute_dict, context_dict

    @classmethod
    def _from_dict(
        cls,
        attribute_dict: Dict[str, Any],
        context_dict: Dict[str, Any],
        loaded_dict: Optional[Dict[str, BaseSimObj]] = None,
    ) -> Tuple[BaseSimObj, Dict[str, BaseSimObj]]:
        battery_list = []
        for ev in attribute_dict["battery_list"]:
            # noinspection PyProtectedMember
            ev_elt, loaded_dict = BaseSimObj._build_from_id(
                ev, context_dict=context_dict, loaded_dict=loaded_dict
            )
            battery_list.append(ev_elt)
        out_obj = cls(attribute_dict["timestamp"], battery_list)
        return out_obj, loaded_dict



from .. import ChargingNetwork
from ..current import Current
from ...models.evse import get_evse_by_type
import numpy as np


def _register_evses(network, evse_ids, evse_type, voltage, angle):
    for evse_id in evse_ids:
        network.register_evse(get_evse_by_type(evse_id, evse_type), voltage, angle)


def _add_line2line_evses(
    network, ab, bc, ca, voltage, evse_type, phi_ab=30, phi_bc=-90, phi_ca=150
):
    _register_evses(network, ab, evse_type, voltage, phi_ab)
    _register_evses(network, bc, evse_type, voltage, phi_bc)
    _register_evses(network, ca, evse_type, voltage, phi_ca)

    currents = {}
    currents["ab"] = Current(ab)
    currents["bc"] = Current(bc)
    currents["ca"] = Current(ca)

    currents["a"] = currents["ab"] - currents["ca"]
    currents["b"] = currents["bc"] - currents["ab"]
    currents["c"] = currents["ca"] - currents["bc"]

    return currents


def jpl_acn(
    basic_evse=False,
    voltage=208,
    first_transformer_cap=45,
    third_fourth_transformer_cap=150,
    network_type=ChargingNetwork,
):
    """ Predefined ChargingNetwork for the JPL ACN.

    Args:
        basic_evse (bool): If True use BASIC EVSE type instead of actual AeroViroment types.
        voltage (float): Default voltage at the EVSEs. Does not affect the current rating of the transformer which is
            based on nominal voltages in the network. 277V LL for delta primary and 120V LN for wye secondary. [V]
        first_transformer_cap (float): Capacity of transformer feeding the 1st floor. Default: 45. [kW]
        third_fourth_transformer_cap (float): Capacity of transformer feeding the 3rd and 4th floors. Default: 150. [kW]
        network_type (ChargingNetwork like): Type to be returned. Should be ChargingNetwork or a subclass.

    Returns:
        ChargingNetwork: A ChargingNetwork-like object configured with the EVSEs and constraints of the JPL ACN.
    """
    network = network_type()

    if basic_evse:
        evse_type = "BASIC"
    else:
        evse_type = "AeroVironment"

    def _delta_wye_transformer(name, Isec, cap, secondary_voltage=120, n=4):
        Ipa = (1 / n) * (Isec["a"] - Isec["c"])
        Ipb = (1 / n) * (Isec["b"] - Isec["a"])
        Ipc = (1 / n) * (Isec["c"] - Isec["b"])

        primary_side_constr = cap * 1000 / 3 / secondary_voltage * np.sqrt(3)
        secondary_side_constr = cap * 1000 / 3 / secondary_voltage
        network.add_constraint(
            Isec["a"], secondary_side_constr, name="{0} Secondary A".format(name)
        )
        network.add_constraint(
            Isec["b"], secondary_side_constr, name="{0} Secondary B".format(name)
        )
        network.add_constraint(
            Isec["c"], secondary_side_constr, name="{0} Secondary C".format(name)
        )
        network.add_constraint(
            Ipa, primary_side_constr, name="{0} Primary A".format(name)
        )
        network.add_constraint(
            Ipb, primary_side_constr, name="{0} Primary B".format(name)
        )
        network.add_constraint(
            Ipc, primary_side_constr, name="{0} Primary C".format(name)
        )

    # -------- 1st Floor 45 kW Transformer -----------------
    # Sub-panel 1 (Max 100 A / phase)
    first_floor_sp1 = _add_line2line_evses(
        network, ["AG-1F12", "AG-1F14"], [], ["AG-1F11", "AG-1F13"], voltage, evse_type
    )

    # Sub-panel 2 (Max 100 A / phase)
    first_floor_sp2 = _add_line2line_evses(
        network,
        ["AG-1F03", "AG-1F06"],
        ["AG-1F01", "AG-1F04"],
        ["AG-1F02", "AG-1F05"],
        voltage,
        evse_type,
    )

    # Additional EVSEs on main panel
    add_first_floor = _add_line2line_evses(
        network, ["AG-1F10"], ["AG-1F07", "AG-1F09"], ["AG-1F08"], voltage, evse_type
    )
    first_floor_transformer = dict()
    for p in "abc":
        first_floor_transformer[p] = (
            add_first_floor[p] + first_floor_sp1[p] + first_floor_sp2[p]
        )

    # -------- 3rd and 4th Floors 150 kW Transformer -----------------

    # 3rd Floor (Max 225 A / phase)
    third_floor_panel = _add_line2line_evses(
        network,
        [
            "AG-3F16",
            "AG-3F17",
            "AG-3F20",
            "AG-3F23",
            "AG-3F25",
            "AG-3F26",
            "AG-3F29",
            "AG-3F33",
        ],
        ["AG-3F18", "AG-3F21", "AG-3F27", "AG-3F30", "AG-3F31"],
        ["AG-3F15", "AG-3F19", "AG-3F22", "AG-3F24", "AG-3F28", "AG-3F32"],
        voltage,
        evse_type,
    )

    # 4th Floor (Max 225 A / phase)
    fourth_floor_panel = _add_line2line_evses(
        network,
        [
            "AG-4F35",
            "AG-4F36",
            "AG-4F39",
            "AG-4F42",
            "AG-4F44",
            "AG-4F45",
            "AG-4F48",
            "AG-4F52",
        ],
        ["AG-4F37", "AG-4F40", "AG-4F46", "AG-4F49", "AG-4F50"],
        ["AG-4F34", "AG-4F38", "AG-4F41", "AG-4F43", "AG-4F47", "AG-4F51"],
        voltage,
        evse_type,
    )
    third_fourth_transformer = {
        p: third_floor_panel[p] + fourth_floor_panel[p] for p in "abc"
    }

    # ------ Add Constraints -----------------
    # Line Constraints
    for p in "abc":
        network.add_constraint(
            first_floor_sp1[p], 100, name="First Floor SP1 I_{0}".format(p)
        )
        network.add_constraint(
            first_floor_sp2[p], 100, name="First Floor SP2 I_{0}".format(p)
        )
        network.add_constraint(
            third_floor_panel[p], 225, name="Third Floor Panel I_{0}".format(p)
        )
        network.add_constraint(
            fourth_floor_panel[p], 225, name="Fourth Floor Panel I_{0}".format(p)
        )

    # Transformer Constraints
    _delta_wye_transformer(
        "First Floor Transformer", first_floor_transformer, first_transformer_cap
    )
    _delta_wye_transformer(
        "Third/Fourth Floor Transformer",
        third_fourth_transformer,
        third_fourth_transformer_cap,
    )

    return network
